from refinitiv.data.content.ipa.curves._bond_curves import curves
from tests.unit.conftest import StubResponse, StubSession


def test_search_definition():
    # given
    response = StubResponse(
        {
            "data": [
                {
                    "curveParameters": {
                        "calculateParYield": False,
                        "priceSide": "Mid",
                        "calendarAdjustment": "Calendar",
                        "calendars": ["EMU_FI"],
                        "calibrationModel": "BasisSpline",
                        "useDurationWeightedMinimization": False,
                        "returnCalibratedParameters": False,
                        "valuationDate": "2022-12-08",
                        "useDelayedDataIfDenied": False,
                    },
                    "curveDefinition": {
                        "name": "Eurozone GOV Par Benchmark Curve",
                        "issuerType": "Sovereign",
                        "country": "EP",
                        "source": "Refinitiv",
                        "currency": "EUR",
                        "curveSubType": "GovernmentBenchmark",
                        "id": "b302758b-1049-49b5-ba3b-da726712e133",
                    },
                    "curvePoints": [
                        {
                            "endDate": "2023-03-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.9954433125749454,
                            "ratePercent": 1.8694729555234435,
                            "tenor": "3M",
                        },
                        {
                            "endDate": "2023-06-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.9908068253445579,
                            "ratePercent": 1.8694729555234213,
                            "tenor": "6M",
                        },
                        {
                            "endDate": "2023-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.981648349586145,
                            "ratePercent": 1.8694729555234213,
                            "tenor": "1Y",
                        },
                        {
                            "endDate": "2024-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.9636822499969483,
                            "ratePercent": 1.8668953382534603,
                            "tenor": "2Y",
                        },
                        {
                            "endDate": "2025-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.9459012159432262,
                            "ratePercent": 1.8711961113877074,
                            "tenor": "3Y",
                        },
                        {
                            "endDate": "2026-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.9285423675021957,
                            "ratePercent": 1.8707653196890606,
                            "tenor": "4Y",
                        },
                        {
                            "endDate": "2027-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.9115020825793422,
                            "ratePercent": 1.8705068455443064,
                            "tenor": "5Y",
                        },
                        {
                            "endDate": "2028-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.8950010766357648,
                            "ratePercent": 1.866036146990635,
                            "tenor": "6Y",
                        },
                        {
                            "endDate": "2029-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.8783093544751289,
                            "ratePercent": 1.870949944479694,
                            "tenor": "7Y",
                        },
                        {
                            "endDate": "2030-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.8621909282465827,
                            "ratePercent": 1.8707653196890606,
                            "tenor": "8Y",
                        },
                        {
                            "endDate": "2031-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.8463683017414042,
                            "ratePercent": 1.87062172286101,
                            "tenor": "9Y",
                        },
                        {
                            "endDate": "2032-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.8312147776735689,
                            "ratePercent": 1.86586430960769,
                            "tenor": "10Y",
                        },
                        {
                            "endDate": "2034-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.8005808057529976,
                            "ratePercent": 1.8707653196890606,
                            "tenor": "12Y",
                        },
                        {
                            "endDate": "2037-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.7572703576225682,
                            "ratePercent": 1.870851477883062,
                            "tenor": "15Y",
                        },
                        {
                            "endDate": "2042-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.6902535080485741,
                            "ratePercent": 1.8707653196890606,
                            "tenor": "20Y",
                        },
                        {
                            "endDate": "2047-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.629167510093972,
                            "ratePercent": 1.8707136248076406,
                            "tenor": "25Y",
                        },
                        {
                            "endDate": "2052-12-12",
                            "startDate": "2022-12-12",
                            "discountFactor": 0.5743303216160913,
                            "ratePercent": 1.8656924725146018,
                            "tenor": "30Y",
                        },
                    ],
                    "constituents": {
                        "creditInstruments": {
                            "EUR": {
                                "bonds": [
                                    {
                                        "fields": {
                                            "bid": {"value": 99.407},
                                            "ask": {"value": 99.507},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU10YT=RR",
                                            "fixedRatePercent": 1.7,
                                            "endDate": "2032-08-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:08JUL2022 DMC:F EMC:S FRCD:15AUG2023 FRQ:1 ISSUE:08JUL2022 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 88.906},
                                            "ask": {"value": 89.065},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU15YT=RR",
                                            "fixedRatePercent": 1.0,
                                            "endDate": "2038-05-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:29APR2022 DMC:F EMC:S FRCD:15MAY2023 FRQ:1 ISSUE:29APR2022 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 99.983},
                                            "ask": {"value": 99.985},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU1MT=RR",
                                            "fixedRatePercent": 0.0,
                                            "endDate": "2022-12-14",
                                            "template": "ACC:A0 CFADJ:NO CLDR:EMU_FI DMC:F EMC:S ISSUE:15DEC2021 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 97.951},
                                            "ask": {"value": 98.001},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU1YT=RR",
                                            "fixedRatePercent": 0.0,
                                            "endDate": "2023-11-22",
                                            "template": "ACC:A0 CFADJ:NO CLDR:EMU_FI DMC:F EMC:S ISSUE:23NOV2022 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 124.449},
                                            "ask": {"value": 124.811},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU20YT=RR",
                                            "fixedRatePercent": 3.25,
                                            "endDate": "2042-07-04",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:04JUL2010 DMC:F EMC:S FRCD:04JUL2011 FRQ:1 ISSUE:23JUL2010 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 116.551},
                                            "ask": {"value": 116.754},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU25YT=RR",
                                            "fixedRatePercent": 2.5,
                                            "endDate": "2046-08-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:28FEB2014 DMC:F EMC:S FRCD:15AUG2015 FRQ:1 ISSUE:28FEB2014 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 100.401},
                                            "ask": {"value": 100.521},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU2YT=RR",
                                            "fixedRatePercent": 2.2,
                                            "endDate": "2024-12-12",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:10NOV2022 DMC:F EMC:S FRCD:12DEC2023 FRQ:1 ISSUE:10NOV2022 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 63.304},
                                            "ask": {"value": 63.468},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU30YT=RR",
                                            "fixedRatePercent": 0.0,
                                            "endDate": "2052-08-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:15AUG2021 DMC:F EMC:S FRCD:15AUG2022 FRQ:1 ISSUE:08SEP2021 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 99.431},
                                            "ask": {"value": 99.461},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU3MT=RR",
                                            "fixedRatePercent": 0.0,
                                            "endDate": "2023-04-13",
                                            "template": "ACC:A0 CFADJ:NO CLDR:EMU_FI DMC:F EMC:S ISSUE:13APR2022 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 97.084},
                                            "ask": {"value": 97.204},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU3YT=RR",
                                            "fixedRatePercent": 0.5,
                                            "endDate": "2025-02-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:16JAN2015 DMC:F EMC:S FRCD:15FEB2016 FRQ:1 ISSUE:16JAN2015 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 96.014},
                                            "ask": {"value": 96.134},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU4YT=RR",
                                            "fixedRatePercent": 0.5,
                                            "endDate": "2026-02-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:15JAN2016 DMC:F EMC:S FRCD:15FEB2017 FRQ:1 ISSUE:15JAN2016 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 97.875},
                                            "ask": {"value": 97.967},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU5YT=RR",
                                            "fixedRatePercent": 1.3,
                                            "endDate": "2027-10-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:30JUN2022 DMC:F EMC:S FRCD:15OCT2023 FRQ:1 ISSUE:30JUN2022 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 98.969},
                                            "ask": {"value": 99.014},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU6MT=RR",
                                            "fixedRatePercent": 0.0,
                                            "endDate": "2023-06-21",
                                            "template": "ACC:A0 CFADJ:NO CLDR:EMU_FI DMC:F EMC:S ISSUE:22JUN2022 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 93.944},
                                            "ask": {"value": 94.019},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU6YT=RR",
                                            "fixedRatePercent": 0.5,
                                            "endDate": "2028-02-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:12JAN2018 DMC:F EMC:S FRCD:15FEB2019 FRQ:1 ISSUE:12JAN2018 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 102.046},
                                            "ask": {"value": 102.306},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU7YT=RR",
                                            "fixedRatePercent": 2.1,
                                            "endDate": "2029-11-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:20OCT2022 DMC:F EMC:S FRCD:15NOV2023 FRQ:1 ISSUE:20OCT2022 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 88.543},
                                            "ask": {"value": 88.663},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU8YT=RR",
                                            "fixedRatePercent": 0.0,
                                            "endDate": "2030-02-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:10JAN2020 DMC:F EMC:S FRCD:15FEB2021 FRQ:1 ISSUE:10JAN2020 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 98.381},
                                            "ask": {"value": 98.444},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU9MT=RR",
                                            "fixedRatePercent": 0.0,
                                            "endDate": "2023-09-20",
                                            "template": "ACC:A0 CFADJ:NO CLDR:EMU_FI DMC:F EMC:S ISSUE:21SEP2022 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                    {
                                        "fields": {
                                            "bid": {"value": 86.954},
                                            "ask": {"value": 87.104},
                                        },
                                        "instrumentDefinition": {
                                            "instrumentCode": "EU9YT=RR",
                                            "fixedRatePercent": 0.0,
                                            "endDate": "2031-02-15",
                                            "template": "ACC:AA CCM:BBAA CFADJ:NO CLDR:EMU_FI DATED:08JAN2021 DMC:F EMC:S FRCD:15FEB2022 FRQ:1 ISSUE:08JAN2021 NOTIONAL:1 PX:C PXRND:1E-5:NEAR REFDATE:MATURITY RP:1 SETTLE:2WD XD:NO",
                                            "quotationMode": "PercentCleanPrice",
                                        },
                                    },
                                ]
                            }
                        }
                    },
                }
            ]
        }
    )
    session = StubSession(is_open=True, response=response)
    definition = curves.Definition(
        curve_definition=curves.CreditCurveDefinition(
            reference_entity="0#EUGOVPBMK=R",
            reference_entity_type=curves.ReferenceEntityType.CHAIN_RIC,
        )
    )

    # when
    response = definition.get_data(session)

    # then
    assert response.data.raw
    assert not response.data.df.empty



import os

import pytest

from refinitiv.data import _configure as configure
from .conftest import (
    remove_user_config,
    write_user_config,
    expected_default_user_config_path,
    expected_default_project_config_path,
    remove_project_config,
    write_project_config,
    get_user_config_path,
    get_project_config_path,
    expected_default_config_filename,
    RDPLIB_ENV_DIR,
    RD_LIB_CONFIG_PATH,
    get_client_config_path,
    annotate,
)


def test_number_config_files_paths_by_default_is_2():
    # given

    # when
    configure.reload()

    # then
    assert len(configure._config_files_paths) == 2


def test_default_config_filename():
    # given

    # when
    configure.reload()

    # then
    assert configure._default_config_file_name == expected_default_config_filename


def test_user_config_file_not_exists_but_has_in_config_files_paths():
    # given
    remove_user_config()

    # when
    configure.reload()

    # then
    assert get_user_config_path(configure) == expected_default_user_config_path


def test_user_config_file_exists():
    # given
    write_user_config('{"prop": "value"}')

    # when
    configure.reload()

    # then
    assert get_user_config_path(configure) == expected_default_user_config_path


def test_user_config_file_created_after_module_imported():
    # given
    configure.reload()

    # when
    write_user_config('{"prop": "value"}')

    # then
    assert get_user_config_path(configure) == expected_default_user_config_path


def test_project_config_file_not_exists_but_has_in_config_files_paths():
    # given
    remove_project_config()

    # when
    configure.reload()

    # then
    assert get_project_config_path(configure) == expected_default_project_config_path


def test_project_config_file_exists():
    # given
    write_project_config('{"prop":"value"}')

    # when
    configure.reload()

    # then
    assert get_project_config_path(configure) == expected_default_project_config_path


def test_project_config_file_created_after_module_imported():
    # given
    configure.reload()

    # when
    write_user_config('{"prop": "value"}')

    # then
    assert get_project_config_path(configure) == expected_default_project_config_path


def test_set_RDPLIB_ENV_DIR(monkeypatch, tmpdir):
    # given
    testing_env_dir = "test_env_dir"
    monkeypatch.setenv(RDPLIB_ENV_DIR, testing_env_dir)

    # when
    configure.reload()

    # then
    assert testing_env_dir in get_project_config_path(configure)


def test_set_RDPLIB_ENV_and_set_RDPLIB_ENV_DIR(monkeypatch, tmpdir):
    #  given
    testing_env_dir = "test_env_dir"
    monkeypatch.setenv(RDPLIB_ENV_DIR, testing_env_dir)
    # when
    configure.reload()
    # then
    assert testing_env_dir in get_project_config_path(configure)


@pytest.mark.parametrize("testing_path", ["/test/path/to/", "\\test\\path\\to\\"])
def test_set_RDP_LIB_CONFIG(monkeypatch, testing_path):
    # given
    monkeypatch.setenv(RD_LIB_CONFIG_PATH, testing_path)

    # when
    configure.reload()

    # then
    assert get_client_config_path(configure) == os.path.join(
        testing_path, "refinitiv-data.config.json"
    )


@pytest.mark.parametrize(
    "testing_rootdir, testing_filename, expected_filepath",
    [annotate(rootdir=None, filename=None, expected_filepath=None)],
)
def test_get_filepath_function(testing_rootdir, testing_filename, expected_filepath):
    # given
    configure.reload()

    # when
    testing_filepath = configure._get_filepath(testing_rootdir, testing_filename)

    # then
    assert testing_filepath == expected_filepath



from typing import Union, TYPE_CHECKING

from ...._curves._cross_currency_curves._definitions._create import CreateRequest
from ...._curves._cross_currency_curves._definitions._delete import DeleteRequest
from ...._curves._cross_currency_curves._definitions._get import (
    CrossCurrencyCurveDefinitionKeys,
    GetRequest,
)
from ...._curves._cross_currency_curves._definitions._update import UpdateRequest

from ......_content_type import ContentType
from ......delivery._data import RequestMethod
from ......delivery._data._data_provider_layer import DataProviderLayer

if TYPE_CHECKING:
    from ...._curves._cross_currency_curves._definitions._types import (
        OptMainConstituentAssetClass,
        OptRiskType,
        Segments,
        OptTurns,
        CurveCreateDefinition,
        OptOverrides,
        CurveUpdateDefinition,
    )
    from ......delivery._data._data_provider import Response
    from ......_core.session import Session
    from ......_types import ExtendedParams, OptStr, OptBool


UnionSaveRequest = Union[CreateRequest, UpdateRequest]


def delete(id: str, session: "Session" = None) -> "Response":
    """
    Delete the Cross Currency curve definition for the definition Id provided.

    Parameters
    ----------
    id : str
        The identifier of the cross currency definition.
    session : Session, optional
        session=None. Means default session would be used.

    Examples
    --------
    >>> from refinitiv.data.content.ipa.curves._cross_currency_curves import definitions
    >>> response = definitions.manage.delete(id="334b89f6-e272-4900-ad1e-dfefv")
    """
    request_item = DeleteRequest(id)
    data_provider_layer = DataProviderLayer(
        data_type=ContentType.CROSS_CURRENCY_CURVES_DEFINITIONS_DELETE,
        request_items=request_item,
        method=RequestMethod.POST,
    )
    return data_provider_layer.get_data(session)


def get(
    id: "OptStr" = None,
    main_constituent_asset_class: "OptMainConstituentAssetClass" = None,
    risk_type: "OptRiskType" = None,
    base_currency: "OptStr" = None,
    base_index_name: "OptStr" = None,
    is_non_deliverable: "OptBool" = None,
    name: "OptStr" = None,
    quoted_currency: "OptStr" = None,
    quoted_index_name: "OptStr" = None,
    source: "OptStr" = None,
    extended_params: "ExtendedParams" = None,
    session: "Session" = None,
) -> "Response":
    """
    Gets the Commodities curve definition for the definition provided.

    Parameters
    ----------
    id : str, optional
        The identifier of the cross currency definitions.
    main_constituent_asset_class : MainConstituentAssetClass, optional
        The asset class used to generate the zero coupon curve. the possible values are:
        * fxforward   * swap
    risk_type : RiskType, optional
        The risk type to which the generated cross currency curve is sensitive. the
        possible value is:   * 'crosscurrency'
    base_currency : str, optional
        The base currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'eur').
    base_index_name : str, optional
        The name of the floating rate index (e.g., 'estr') applied to the base currency.
    is_non_deliverable : bool, optional
        An indicator whether the instrument is non-deliverable:   * true: the instrument
        is non-deliverable,   * false: the instrument is not non-deliverable. the
        property can be used to retrieve cross currency definition for the adjusted
        interest rate curve.
    name : str, optional
        The fxcross currency pair applied to the reference or pivot currency. it is
        expressed in iso 4217 alphabetical format (e.g., 'eur usd fxcross').
    quoted_currency : str, optional
        The quoted currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'usd').
    quoted_index_name : str, optional
        The name of the floating rate index (e.g., 'sofr') applied to the quoted
        currency.
    source : str, optional
        A user-defined string that is provided by the creator of a curve. curves created
        by refinitiv have the 'refinitiv' source.
    extended_params : ExtendedParams, optional
        If necessary other parameters.
    session : Session, optional
        session=None - means default session would be used.

    Examples
    --------
    >>> from refinitiv.data.content.ipa.curves._cross_currency_curves import definitions
    >>> response = definitions.manage.get(
    ...     base_currency="EUR",
    ...     base_index_name="EURIBOR",
    ...     quoted_currency="USD",
    ...     quoted_index_name="LIBOR",
    >>> )
    """
    definition_obj = CrossCurrencyCurveDefinitionKeys(
        main_constituent_asset_class=main_constituent_asset_class,
        risk_type=risk_type,
        base_currency=base_currency,
        base_index_name=base_index_name,
        id=id,
        is_non_deliverable=is_non_deliverable,
        name=name,
        quoted_currency=quoted_currency,
        quoted_index_name=quoted_index_name,
        source=source,
    )
    request_item = GetRequest(curve_definition=definition_obj)
    content_provider_layer = DataProviderLayer(
        data_type=ContentType.CROSS_CURRENCY_CURVES_DEFINITIONS_GET,
        request_items=request_item,
        extended_params=extended_params,
    )
    response = content_provider_layer.get_data(session=session)
    return response


def create(
    curve_definition: "CurveCreateDefinition" = None,
    overrides: "OptOverrides" = None,
    segments: "Segments" = None,
    turns: "OptTurns" = None,
    extended_params: "ExtendedParams" = None,
    session: "Session" = None,
) -> "Response":
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    curve_definition : CrossCurrencyCurveDefinitionDescription
        CrossCurrencyCurveDefinitionDescription object
    overrides : list of OverrideBidAsk, optional
        OverrideBidAsk object.
    segments : list of CrossCurrencyInstrumentsSegment
        list of CrossCurrencyInstrumentsSegment objects
    turns : list of OverrideFxForwardTurn, optional
    extended_params : ExtendedParams, optional
        If necessary other parameters.
    session : Session, optional
        session=None - means default session would be used

    Examples
    --------
    >>> from refinitiv.data.content.ipa.curves._cross_currency_curves import definitions
    >>> response = definitions.manage.create(
    ...     curve_definition=definitions.CrossCurrencyCurveDefinitionDescription(
    ...         source="SourceName",
    ...         name="Name of the Curve854",
    ...         base_currency="EUR",
    ...         base_index_name="ESTR",
    ...         quoted_currency="USD",
    ...         quoted_index_name="SOFR",
    ...         is_non_deliverable=False
    ...     ),
    ...     segments=[
    ...         definitions.CrossCurrencyInstrumentsSegment(
    ...             start_date="2021-01-01",
    ...             constituents=definitions.CrossCurrencyConstituentsDescription(
    ...                 fx_forwards=[
    ...                     definitions.FxForwardInstrumentDescription(
    ...                         instrument_definition=definitions.FxForwardInstrumentDefinition(
    ...                             instrument_code="EUR1M=",
    ...                             tenor="1M",
    ...                             is_non_deliverable=False,
    ...                             quotation_mode=definitions.QuotationMode.OUTRIGHT
    ...                         )
    ...                     )
    ...                 ]
    ...             ),
    ...         )
    ...     ]
    >>> )
    """
    request_items = CreateRequest(
        curve_definition=curve_definition,
        overrides=overrides,
        segments=segments,
        turns=turns,
    )

    response = _save(
        ContentType.CROSS_CURRENCY_CURVES_DEFINITIONS_CREATE,
        request_items=request_items,
        extended_params=extended_params,
        session=session,
    )
    return response


def update(
    curve_definition: "CurveUpdateDefinition" = None,
    overrides: "OptOverrides" = None,
    segments: "Segments" = None,
    turns: "OptTurns" = None,
    extended_params: "ExtendedParams" = None,
    session: "Session" = None,
) -> "Response":
    """
    Updates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    curve_definition : CrossCurrencyCurveUpdateDefinition
        CrossCurrencyCurveUpdateDefinition object.
    overrides : list of OverrideBidAsk, optional
        OverrideBidAsk object.
    segments : list of CrossCurrencyInstrumentsSegment
        list of CrossCurrencyInstrumentsSegment objects
    turns : list of OverrideFxForwardTurn, optional
    extended_params : ExtendedParams, optional
        If necessary other parameters.
    session : Session, optional
        session=None - means default session would be used

    Examples
    --------
    >>> from refinitiv.data.content.ipa.curves._cross_currency_curves import definitions
    >>> response = definitions.manage.update(
    ...     curve_definition=definitions.CrossCurrencyCurveUpdateDefinition(
    ...         id="7bdb00f3-0a48-40be-ace2-6d3cfd0e8e59",
    ...         source="SourceName",
    ...         name="rename curve",
    ...         base_currency="EUR",
    ...         base_index_name="ESTR",
    ...         quoted_currency="USD",
    ...         quoted_index_name="SOFR",
    ...         is_non_deliverable=False
    ...     ),
    ...     segments=[
    ...         definitions.CrossCurrencyInstrumentsSegment(
    ...             start_date="2021-01-01",
    ...             constituents=definitions.CrossCurrencyConstituentsDescription(
    ...                 fx_forwards=[
    ...                     definitions.FxForwardInstrumentDescription(
    ...                         instrument_definition=definitions.FxForwardInstrumentDefinition(
    ...                             instrument_code="EUR1M=",
    ...                             tenor="1M",
    ...                             is_non_deliverable=False,
    ...                             quotation_mode=definitions.QuotationMode.OUTRIGHT
    ...                         )
    ...                     )
    ...                 ]
    ...             ),
    ...         )
    ...     ]
    >>> )
    """
    request_items = UpdateRequest(
        curve_definition=curve_definition,
        overrides=overrides,
        segments=segments,
        turns=turns,
    )

    response = _save(
        ContentType.CROSS_CURRENCY_CURVES_DEFINITIONS_UPDATE,
        request_items=request_items,
        extended_params=extended_params,
        session=session,
    )
    return response


def _save(
    content_type: ContentType,
    request_items: UnionSaveRequest,
    extended_params: "ExtendedParams" = None,
    session: "Session" = None,
) -> "Response":
    data_provider_layer = DataProviderLayer(
        data_type=content_type,
        request_items=request_items,
        method=RequestMethod.POST,
        extended_params=extended_params,
    )

    return data_provider_layer.get_data(session=session)



# coding: utf8

from typing import Optional

from ..._instrument_definition import ObjectDefinition
from .._enums import BarrierMode
from ._fx_double_barrier_info import FxDoubleBarrierInfo


class FxDoubleBarrierDefinition(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    barrier_down : FxDoubleBarrierInfo, optional
        Barrier Information for the lower barrier
    barrier_mode : BarrierMode, optional
        Barrier Mode of the double barrier option
    barrier_up : FxDoubleBarrierInfo, optional
        Barrier Information for the upper barrier
    """

    def __init__(
        self,
        barrier_down: Optional[FxDoubleBarrierInfo] = None,
        barrier_mode: Optional[BarrierMode] = None,
        barrier_up: Optional[FxDoubleBarrierInfo] = None,
    ) -> None:
        super().__init__()
        self.barrier_down = barrier_down
        self.barrier_mode = barrier_mode
        self.barrier_up = barrier_up

    @property
    def barrier_down(self):
        """
        Barrier Information for the lower barrier
        :return: object FxDoubleBarrierInfo
        """
        return self._get_object_parameter(FxDoubleBarrierInfo, "barrierDown")

    @barrier_down.setter
    def barrier_down(self, value):
        self._set_object_parameter(FxDoubleBarrierInfo, "barrierDown", value)

    @property
    def barrier_mode(self):
        """
        Barrier Mode of the double barrier option
        :return: enum BarrierMode
        """
        return self._get_enum_parameter(BarrierMode, "barrierMode")

    @barrier_mode.setter
    def barrier_mode(self, value):
        self._set_enum_parameter(BarrierMode, "barrierMode", value)

    @property
    def barrier_up(self):
        """
        Barrier Information for the upper barrier
        :return: object FxDoubleBarrierInfo
        """
        return self._get_object_parameter(FxDoubleBarrierInfo, "barrierUp")

    @barrier_up.setter
    def barrier_up(self, value):
        self._set_object_parameter(FxDoubleBarrierInfo, "barrierUp", value)



import urllib

import pytest

from refinitiv.data._tools import parse_url, ParseResult, ParseResultBytes
from tests.unit.conftest import args


@pytest.mark.parametrize(
    "input_url,expected_result",
    [
        args(
            input="http://1.2.3.4:80",
            expected=ParseResult(
                scheme="http",
                netloc="1.2.3.4:80",
                path="",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="http://1.2.3.4",
            expected=ParseResult(
                scheme="http",
                netloc="1.2.3.4",
                path="",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="://1.2.3.4:80",
            expected=ParseResult(
                scheme="",
                netloc="",
                path="://1.2.3.4:80",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="//1.2.3.4:80",
            expected=ParseResult(
                scheme="",
                netloc="1.2.3.4:80",
                path="",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="1.2.3.4:80",
            expected=ParseResult(
                scheme="",
                netloc="",
                path="1.2.3.4:80",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="//1.2.3.4",
            expected=ParseResult(
                scheme="",
                netloc="1.2.3.4",
                path="",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="scheme://netloc/path;parameters?query#fragment",
            expected=ParseResult(
                scheme="scheme",
                netloc="netloc",
                path="/path;parameters",
                params="",
                query="query",
                fragment="fragment",
            ),
        ),
        args(
            input="http://docs.python.org:80/3/library/urllib.parse.html?highlight=params#url-parsing",
            expected=ParseResult(
                scheme="http",
                netloc="docs.python.org:80",
                path="/3/library/urllib.parse.html",
                params="",
                query="highlight=params",
                fragment="url-parsing",
            ),
        ),
        args(
            input="//www.cwi.nl:80/%7Eguido/Python.html",
            expected=ParseResult(
                scheme="",
                netloc="www.cwi.nl:80",
                path="/%7Eguido/Python.html",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="www.cwi.nl/%7Eguido/Python.html",
            expected=ParseResult(
                scheme="",
                netloc="",
                path="www.cwi.nl/%7Eguido/Python.html",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="help/Python.html",
            expected=ParseResult(
                scheme="",
                netloc="",
                path="help/Python.html",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="abc:123.html",
            expected=ParseResult(
                scheme="abc",
                netloc="",
                path="123.html",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="123.html:abc",
            expected=ParseResult(
                scheme="",
                netloc="",
                path="123.html:abc",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="abc:123/",
            expected=ParseResult(
                scheme="abc",
                netloc="",
                path="123/",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="abc:123",
            expected=ParseResult(
                scheme="abc",
                netloc="",
                path="123",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input="",
            expected=ParseResult(
                scheme="",
                netloc="",
                path="",
                params="",
                query="",
                fragment="",
            ),
        ),
        args(
            input={},
            expected=ParseResultBytes(
                scheme=b"",
                netloc=b"",
                path=b"",
                params=b"",
                query=b"",
                fragment=b"",
            ),
        ),
        args(
            input=None,
            expected=ParseResultBytes(
                scheme=b"",
                netloc=b"",
                path=b"",
                params=b"",
                query=b"",
                fragment=b"",
            ),
        ),
    ],
)
def test_parse_value(input_url, expected_result):
    # when
    testing_result = parse_url(input_url)

    # then
    assert testing_result == expected_result, input_url


def test_result_type():
    # given
    url = "http://1.2.3.4:80"

    # when
    rd_result = parse_url(url)
    urllib_result = urllib.parse.urlparse(url)

    # then
    assert type(rd_result) == type(urllib_result)



from dataclasses import dataclass
from typing import List, Optional, Union

from ..._types import OptDateTime
from ...content.ipa._enums import PeriodType, DayCountBasis
from ...content.ipa.dates_and_calendars.count_periods import Definition


@dataclass
class CountedPeriods:
    count: float
    tenor: str


def count_periods(
    start_date: "OptDateTime" = None,
    end_date: "OptDateTime" = None,
    period_type: Optional[Union[PeriodType, str]] = None,
    calendars: Optional[List[str]] = None,
    currencies: Optional[List[str]] = None,
    day_count_basis: Optional[Union[DayCountBasis, str]] = None,
) -> CountedPeriods:
    """
    With this function you can get the period of time
    between a start date and an end date using one or many calendars.

    Parameters
    ----------
        start_date: str or datetime or timedelta, optional
            Start date of calculation.
        end_date: str or datetime or timedelta, optional
            End date of calculation.
        period_type : PeriodType or str, optional
            The method we chose to count the period of time based on value from PeriodType enumeration.
        calendars: list of str, optional
            Calendars to use the date for working day or weekend.
            Optional if currencies is provided.
        currencies: list of str, optional
            Currencies to use the date for working day or weekend.
            Optional if calendars is provided.
        day_count_basis: DayCountBasis or str, optional
            Day count basis value from DayCountBasis enumeration.

    Returns
    -------
    CountedPeriods
        Counted periods object with count and tenor values.

    Examples
    --------
     >>> import datetime
     >>> import refinitiv.data as rd
     >>> from refinitiv.data import dates_and_calendars
     >>>
     >>> rd.open_session("platform.default")
     >>>
     >>> counted_period = rd.dates_and_calendars.count_periods(
     ...    start_date=datetime.timedelta(-11),
     ...    end_date=datetime.timedelta(-3),
     ...    period_type=dates_and_calendars.PeriodType.WORKING_DAY,
     ...    currencies=["EUR"],
     >>>)
    """

    response = Definition(
        start_date=start_date,
        end_date=end_date,
        period_type=period_type,
        calendars=calendars,
        currencies=currencies,
        day_count_basis=day_count_basis,
    ).get_data()

    response = CountedPeriods(
        response.data.counted_period.count, response.data.counted_period.tenor
    )

    return response



import os
import queue
import threading
import traceback
from collections import defaultdict, Counter
from typing import Callable, Dict, List, Optional, TYPE_CHECKING, Union

import websocket

from .event import StreamCxnEvent
from .rwf.socket import RwfSocketClient
from .stream_cxn_state import StreamCxnState
from .ws.ws_client import WebSocketClient
from ..._core.log_reporter import LogReporter
from ..._core.session.tools import get_delays, is_desktop_session
from ..._tools import DEBUG, CallbackHandler
from ...usage_collection import LogRecord, RecordData, FilterType, get_usage_logger

if TYPE_CHECKING:
    from ._stream_cxn_config_data import StreamCxnConfig
    from ..._core.session import Session

LOGIN_STREAM_ID = 2
MAX_LISTENERS = 2000


class StreamConnection(threading.Thread, LogReporter):
    _listener: Optional[Union[WebSocketClient, RwfSocketClient]] = None
    _num_attempt = 0

    def __init__(
        self,
        connection_id: int,
        name: str,
        session: "Session",
        config: "StreamCxnConfig",
    ) -> None:
        self._id: int = connection_id
        self._session: "Session" = session
        self._config: "StreamCxnConfig" = config

        LogReporter.__init__(self, logger=session.logger())
        threading.Thread.__init__(self, name=f"Thread{name}", daemon=True)

        self._state: StreamCxnState = StreamCxnState.Initial
        self._is_emit_reconnected: bool = False
        self._is_auto_reconnect: bool = self._session.stream_auto_reconnection
        self._start_connecting = threading.Event()
        self._connection_result_ready = threading.Event()
        self._listener_created = threading.Event()
        self._timer = threading.Event()

        self.usage_counter = defaultdict(Counter)
        self._usage_close_event = threading.Event()
        self._usage_report_thread = threading.Thread(
            target=self._usage_callback, args=(10,)
        )
        self._usage_lock = threading.Lock()
        self._dispose_lock = threading.Lock()

        self._emitter = CallbackHandler(max_listeners=MAX_LISTENERS)
        self._msg_queue: Optional[queue.Queue] = None
        self._msg_processor: Optional[threading.Thread] = None
        self._delays = get_delays()

        self._classname = f"[{name}]"

    @property
    def session(self) -> "Session":
        return self._session

    @property
    def id(self) -> int:
        return self._id

    @property
    def subprotocol(self) -> str:
        return ""

    @property
    def state(self) -> StreamCxnState:
        return self._state

    @property
    def is_connecting(self) -> bool:
        return self.state is StreamCxnState.Connecting

    @property
    def is_message_processing(self) -> bool:
        return self.state is StreamCxnState.MessageProcessing

    @property
    def is_disconnecting(self) -> bool:
        return self.state is StreamCxnState.Disconnecting

    @property
    def is_disconnected(self) -> bool:
        return self.state is StreamCxnState.Disconnected

    @property
    def is_disposed(self) -> bool:
        return self.state is StreamCxnState.Disposed

    @property
    def can_reconnect(self) -> bool:
        return self._is_auto_reconnect and self._config.has_available_info()

    def wait_start_connecting(self):
        msg = "clear start_connecting event"
        if self.is_disposed:
            self.debug(f"{self._classname} cannot {msg}, {self.state}")
            return

        self.debug(f"{self._classname} {msg}")
        self._start_connecting.clear()

    def start_connecting(self):
        msg = "set start_connecting event"
        if self.is_disposed:
            self.debug(f"{self._classname} cannot {msg}, {self.state}")
            return

        self.debug(f"{self._classname} {msg}")
        self._start_connecting.set()

    def wait_start(self):
        if self.is_disposed:
            self.debug(f"{self._classname} cannot wait start, {self.state}")
            return

        self.start()
        self._listener_created.wait()

    def wait_connection_result(self):
        if self.is_disposed:
            self.debug(f"{self._classname} cannot wait connection result, {self.state}")
            return

        self._connection_result_ready.wait()

    def run(self) -> None:
        self._msg_queue = queue.Queue()
        self._msg_processor = threading.Thread(
            target=self._process_messages, name=f"Msg-Proc-{self.name}", daemon=True
        )
        self._usage_report_thread.start()
        self._msg_processor.start()

        once = False
        self._start_connecting.set()
        while not once or self.can_reconnect:
            once = True

            if self.is_disposed:
                break

            self.debug(f"{self._classname} wait start connecting")
            self._start_connecting.wait()

            self.debug(f"{self._classname} is connecting")
            self._state = StreamCxnState.Connecting
            self._emitter.emit(StreamCxnEvent.CONNECTING, self)

            if self._config.transport == "tcp":
                self._run_tcp_listener()
            else:
                self._run_websocket_listener()

            if self.is_disposed:
                break

            self._state = StreamCxnState.Disconnected
            self._listener_created.clear()
            not self.can_reconnect and self._connection_result_ready.set()

            if self.can_reconnect:
                self._connection_result_ready.clear()
                self._emitter.emit(StreamCxnEvent.RECONNECTING, self)
                self._start_connecting.wait()
                self._is_emit_reconnected = True

                try:
                    self._config.next_available_info()
                except StopIteration:
                    delay = self._delays.next()
                    self.debug(
                        f"{self._classname} tried all infos, "
                        f"waiting time {delay} secs until the next attempt."
                    )
                    self._timer.wait(delay)
                    self._num_attempt += 1

                url = self._config.url
                self.debug(f"{self._classname} try to reconnect over url {url}")

        self.dispose()

    def _run_tcp_listener(self):
        def on_open(app):
            self._on_ws_open(app)
            self._state = StreamCxnState.MessageProcessing
            self._connection_result_ready.set()

        host, port = self._config.url.split(":")
        port = int(port) if port else None

        cfg = self._session.config
        cfg_prefix = "apis.streaming.pricing"
        field_dict_path = cfg.get(f"{cfg_prefix}.field_dict_path")
        enumtype_path = cfg.get(f"{cfg_prefix}.enumtype_path")

        self.debug(
            f"{self._classname} connect\n"
            f"\tnum_attempt = {self._num_attempt}\n"
            f"\turl = {self._config.url}\n"
            f"\ttransport   = {self._config.transport}\n"
        )

        self._listener = RwfSocketClient(
            host=host,
            port=port,
            field_dict_path=field_dict_path,
            enumtype_path=enumtype_path,
            on_open=on_open,
            on_message=self._on_message,
            on_close=self._on_ws_close,
            session=self._session,
        )
        self._listener_created.set()
        self._listener.run_forever()

    def _run_websocket_listener(self):
        if DEBUG:
            websocket.enableTrace(True)

        cookie = None
        user_id = os.getenv("REFINITIV_AAA_USER_ID")
        if is_desktop_session(self._session) and user_id:
            cookie = f"user-uuid={user_id}"

        headers = ["User-Agent: Python"] + self._config.headers
        subprotocols = [self.subprotocol]

        self.debug(
            f"{self._classname} connect\n"
            f"\tnum_attempt : {self._num_attempt}\n"
            f"\turl         : {self._config.url}\n"
            f"\theaders     : {headers}\n"
            f"\tcookies     : {cookie}\n"
            f"\ttransport   : {self._config.transport}\n"
            f"\tsubprotocols: {subprotocols}"
        )

        self._listener = WebSocketClient(
            url=self._config.url,
            header=headers,
            cookie=cookie,
            on_open=self._on_ws_open,
            on_message=self._on_message,
            on_error=self._on_ws_error,
            on_close=self._on_ws_close,
            on_ping=self._on_ws_ping,
            on_pong=self._on_ws_pong,
            subprotocols=subprotocols,
        )
        self._listener_created.set()
        proxy_config = self._config.proxy_config
        http_proxy_host = None
        http_proxy_port = None
        http_proxy_auth = None
        proxy_type = None
        if proxy_config:
            http_proxy_host = proxy_config.host
            http_proxy_port = proxy_config.port
            http_proxy_auth = proxy_config.auth
            proxy_type = proxy_config.type

        no_proxy = self._config.no_proxy

        self._listener.run_forever(
            http_proxy_host=http_proxy_host,
            http_proxy_port=http_proxy_port,
            http_proxy_auth=http_proxy_auth,
            http_no_proxy=no_proxy,
            proxy_type=proxy_type,
            skip_utf8_validation=True,
        )

    def start_disconnecting(self) -> None:
        if not self.is_message_processing:
            self.debug(f"{self._classname} cannot start disconnecting, {self.state}")
            return

        self.debug(f"{self._classname} is start disconnecting [dis]")
        self._state = StreamCxnState.Disconnecting
        self._connection_result_ready.clear()
        self._start_connecting.clear()
        self._emitter.emit(StreamCxnEvent.DISCONNECTING, self)

    def end_disconnecting(self):
        if self.is_disposed:
            self.debug(f"{self._classname} cannot end disconnecting, {self.state}")
            return

        if self.is_disconnecting:
            self._state = StreamCxnState.Disconnected
            self._start_connecting.clear()
        else:
            self.debug(f"{self._classname} cannot end disconnecting, {self.state}")

    def dispose(self) -> None:
        if self.is_disposed or self.is_message_processing:
            self.debug(f"{self._classname} cannot dispose, {self.state}")
            return

        with self._dispose_lock:
            self.debug(f"{self._classname} is disposing [d], {self.state}")
            self._state = StreamCxnState.Disposed

            close_message = self.get_close_message()
            if close_message and not self.is_connecting:
                self.send_message(close_message)

            self._listener_created.set()
            self._listener_created = None
            self._listener.close()
            self._listener = None

            self._usage_close_event.set()
            self._usage_report_thread.join(5)
            self._msg_queue.put("__dispose__")
            self._msg_queue = None
            self._msg_processor = None
            self._start_connecting = None
            self._connection_result_ready.set()
            self._connection_result_ready = None
            self._emitter.emit(StreamCxnEvent.DISPOSED, self)
            self._emitter = None
            self.debug(f"{self._classname} disposed [D]")

    def get_login_message(self) -> dict:
        # for override
        pass

    def get_close_message(self) -> dict:
        # for override
        pass

    def send_login_message(self) -> None:
        self.send_message(self.get_login_message())

    def send_message(self, message: dict) -> bool:
        if self.is_message_processing:
            self.debug(f"{self._classname} send {message}")
            self._listener.send(message)
            return True
        else:
            self.debug(
                f"{self._classname} cannot send message: "
                f"state={self.state}, message={message}"
            )
            return False

    def on(self, event: str, listener: Callable) -> None:
        if self.is_disposed:
            self.debug(f"{self._classname} cannot on, {self.state}")
            return

        self._emitter.on(event, listener)

    def remove_listener(self, event: str, listener: Callable) -> None:
        if self.is_disposed:
            self.debug(
                f"{self._classname} {self.state}, "
                f"can not remove listener {event}, {listener}"
            )
            return

        self._emitter.remove_listener(event, listener)

    def _on_ws_open(self, ws: websocket.WebSocketApp) -> None:
        self.debug(f"{self._classname} on_ws_open")
        self._delays.reset()
        self._emitter.emit(StreamCxnEvent.CONNECTED, self)
        message = self.get_login_message()
        self.debug(f"{self._classname} send login message {message}")
        self._listener.send(message)

    def _on_message(self, messages: List[Dict]) -> None:
        self.debug(f"{self._classname} on_ws_message {messages}")

        if self.is_connecting:
            if len(messages) > 1:
                raise ValueError(
                    f"Cannot process messages more then one, num={len(messages)}"
                )

            message = messages[0]
            self._handle_login_message(message)

            if self._is_emit_reconnected:
                self.debug(f"Reconnecting is over, emit event Reconnected.")
                self._is_emit_reconnected = False
                self._emitter.emit(StreamCxnEvent.RECONNECTED, self)

        elif self.is_message_processing:
            self._msg_queue.put(messages)

        elif self.is_disconnecting:
            # do nothing and wait for all streams to close
            pass

        else:
            debug_msg = (
                f"{self._classname}._on_ws_message() | "
                f"don't know what to do state={self.state}, "
                f"message={messages}"
            )
            if DEBUG:
                raise ValueError(debug_msg)

            else:
                self.debug(debug_msg)

    def _handle_login_message(self, message: dict):
        # for override
        pass

    def _process_messages(self) -> None:
        while not self.is_disposed:
            messages: List[Dict] = self._msg_queue.get()

            if messages == "__dispose__" and self.is_disposed:
                break

            for message in messages:
                self._process_message(message)

            self._msg_queue.task_done()

    def _process_message(self, message: dict) -> None:
        # for override
        pass

    def _on_ws_close(
        self, ws: websocket.WebSocketApp, close_status_code: str, close_msg: str
    ) -> None:
        self.debug(
            f"{self._classname} on_ws_close: "
            f"close_status_code={close_status_code}, close_msg={close_msg}, "
            f"state={self.state}"
        )
        self._emitter.emit(StreamCxnEvent.DISCONNECTED, self)
        if not self.is_disposed:
            self._listener_created.clear()

    def _on_ws_error(self, ws: websocket.WebSocketApp, exc: Exception) -> None:
        self.debug(f"{self._classname} on_ws_error: Exception: {exc}")
        if DEBUG:
            self.debug(f"{traceback.format_exc()}")

    def _on_ws_ping(self, message: dict) -> None:
        self.debug(f"{self._classname} ping {message}")

    def _on_ws_pong(self, ws, message: dict) -> None:
        self.debug(f"{self._classname} pong {message}")

    def _dump_usage_data(self) -> None:
        if self.usage_counter:
            with self._usage_lock:
                data_to_log = self.usage_counter
                self.usage_counter = defaultdict(Counter)
            get_usage_logger().log(
                LogRecord(
                    name="stream_usage",
                    func_path="stream_usage",
                    data=RecordData(result=data_to_log),
                    filter={FilterType.STREAM},
                )
            )

    def _usage_callback(self, interval: int) -> None:
        while not self._usage_close_event.wait(interval):
            self._dump_usage_data()
        self._dump_usage_data()

    def __str__(self) -> str:
        s = (
            f"{self.__class__.__name__}\n"
            f"\t\tname             : {self.name}\n"
            f"\t\tsubprotocol      : {self.subprotocol}\n"
            f"\t\tis_auto_reconnect: {self._is_auto_reconnect}\n"
            f"\t\tcan_reconnect    : {self.can_reconnect}\n"
            f"\t\tnum_attempt      : {self._num_attempt}"
        )
        return s



from ._discovery_universe import DiscoveryUniverse


class Peers(DiscoveryUniverse):
    """
    Class to get data from peers function.

    Parameters
    ----------
    expression : str
        peers expression


    Examples
    --------
    >>> peers = Peers("VOD.L")
    >>> print(list(peers))
    """

    def __init__(self, expression):
        super().__init__(f"peers({expression})")



from refinitiv.data.content.ipa._curves._bond_curves._curves import (
    CreditCurveParameters,
)
from refinitiv.data.content.ipa._curves._bond_curves._enums import (
    InterestCalculationMethod,
    BasisSplineSmoothModel,
    CalendarAdjustment,
    CalibrationModel,
)
from refinitiv.data.content.ipa._curves._enums import CompoundingType
from refinitiv.data.content.ipa._enums._extrapolation_mode import ExtrapolationMode
from refinitiv.data.content.ipa._curves._bond_curves._enums import (
    InterpolationMode,
    PriceSide,
)


def test_credit_curve_parameters():
    # given
    expected_dict = {
        "interestCalculationMethod": "Dcb_30E_360_ISMA",
        "basisSplineSmoothModel": "AndersonSmoothingSplineModel",
        "calendarAdjustment": "Calendar",
        "calendars": ["string1", "string2"],
        "calibrationModel": "BasisSpline",
        "compoundingType": "Compounded",
        "extrapolationMode": "Constant",
        "interpolationMode": "AkimaMethod",
        "priceSide": "Ask",
        "basisSplineKnots": 10,
        "returnCalibratedParameters": True,
        "useDelayedDataIfDenied": True,
        "useDurationWeightedMinimization": True,
        "useMultiDimensionalSolver": True,
        "valuationDate": "string",
    }

    # when
    testing_obj = CreditCurveParameters(
        interest_calculation_method=InterestCalculationMethod.DCB_30_E_360_ISMA,
        basis_spline_smooth_model=BasisSplineSmoothModel.ANDERSON_SMOOTHING_SPLINE_MODEL,
        calendar_adjustment=CalendarAdjustment.CALENDAR,
        calendars=["string1", "string2"],
        calibration_model=CalibrationModel.BASIS_SPLINE,
        compounding_type=CompoundingType.COMPOUNDED,
        extrapolation_mode=ExtrapolationMode.CONSTANT,
        interpolation_mode=InterpolationMode.AKIMA_METHOD,
        price_side=PriceSide.ASK,
        basis_spline_knots=10,
        return_calibrated_parameters=True,
        use_delayed_data_if_denied=True,
        use_duration_weighted_minimization=True,
        use_multi_dimensional_solver=True,
        valuation_date="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert (
        testing_obj.interest_calculation_method
        == InterestCalculationMethod.DCB_30_E_360_ISMA
    )
    assert (
        testing_obj.basis_spline_smooth_model
        == BasisSplineSmoothModel.ANDERSON_SMOOTHING_SPLINE_MODEL
    )
    assert testing_obj.calendar_adjustment == CalendarAdjustment.CALENDAR
    assert testing_obj.calendars == ["string1", "string2"]
    assert testing_obj.calibration_model == CalibrationModel.BASIS_SPLINE
    assert testing_obj.compounding_type == CompoundingType.COMPOUNDED
    assert testing_obj.extrapolation_mode == ExtrapolationMode.CONSTANT
    assert testing_obj.interpolation_mode == InterpolationMode.AKIMA_METHOD
    assert testing_obj.price_side == PriceSide.ASK
    assert testing_obj.basis_spline_knots == 10
    assert testing_obj.return_calibrated_parameters == True
    assert testing_obj.use_delayed_data_if_denied == True
    assert testing_obj.use_duration_weighted_minimization == True
    assert testing_obj.use_multi_dimensional_solver == True
    assert testing_obj.valuation_date == "string"



import numpy

from refinitiv.data.content.ipa.dates_and_calendars import date_schedule
from refinitiv.data.content.ipa.dates_and_calendars._request_factory import (
    DatesAndCalendarsResponseFactory,
)
from refinitiv.data.content.ipa.dates_and_calendars.date_schedule._date_schedule_data_provider import (
    DateSchedule,
)
from refinitiv.data.delivery._data._data_provider import ParsedData


CONVERTED_DATES = [
    numpy.datetime64("2019-05-07"),
    numpy.datetime64("2019-05-14"),
    numpy.datetime64("2019-05-21"),
    numpy.datetime64("2019-05-28"),
    numpy.datetime64("2019-06-04"),
    numpy.datetime64("2019-06-11"),
    numpy.datetime64("2019-06-18"),
    numpy.datetime64("2019-06-25"),
    numpy.datetime64("2019-07-02"),
    numpy.datetime64("2019-07-09"),
]

MOCKED_DATE_SCHEDULE_DATA = ParsedData(
    {"http_status_code": 200, "http_reason": "OK"},
    {},
    {
        "dates": [
            "2019-05-07",
            "2019-05-14",
            "2019-05-21",
            "2019-05-28",
            "2019-06-04",
            "2019-06-11",
            "2019-06-18",
            "2019-06-25",
            "2019-07-02",
            "2019-07-09",
        ]
    },
)


def test_date_schedule():
    definition = date_schedule.Definition(
        start_date="2019-04-30",
        count=10,
        frequency="Weekly",
        calendars=["EMU", "GER"],
        currencies=["EUR"],
        day_of_week=date_schedule.DayOfWeek.TUESDAY,
    )

    assert (
        definition.request_item.frequency == date_schedule.DateScheduleFrequency.WEEKLY
    )
    assert definition.request_item.start_date == "2019-04-30"
    assert definition.request_item.count == 10
    assert definition.request_item.calendars == ["EMU", "GER"]
    assert definition.request_item.currencies == ["EUR"]
    assert definition.request_item.day_of_week == date_schedule.DayOfWeek.TUESDAY


def test_holidays_create_success():
    response_factory = DatesAndCalendarsResponseFactory(data_class=DateSchedule)
    response = response_factory.create_success(MOCKED_DATE_SCHEDULE_DATA)
    assert not response.data.df.empty
    assert response.data.raw == MOCKED_DATE_SCHEDULE_DATA.content_data
    assert response.data.dates == CONVERTED_DATES



# coding: utf8


from enum import Enum, unique


@unique
class Method(Enum):
    AMERICAN_MONTE_CARLO = "AmericanMonteCarlo"
    ANALYTIC = "Analytic"
    MONTE_CARLO = "MonteCarlo"
    PDE = "PDE"
    TREE = "Tree"



import pytest

from refinitiv.data.content.filings import retrieval


def test_retrieval_definition(create_filings_definition):
    retrieval.Definition(**create_filings_definition)


def test_retrieval_definition_error(create_filings_definition_with_error):
    try:
        retrieval.Definition(**create_filings_definition_with_error)
    except ValueError:
        assert True
    except Exception as e:
        assert False, str(e)


@pytest.mark.parametrize(
    "filings_definition, expected_repr",
    [
        (
            retrieval.Definition(filename="filename_1"),
            "<refinitiv.data.content.filings.retrieval.Definition object at {0} {{filename='filename_1', dcn='None', doc_id='None', filing_id='None'}}>",
        ),
        (
            retrieval.Definition(dcn="dcn_1"),
            "<refinitiv.data.content.filings.retrieval.Definition object at {0} {{filename='None', dcn='dcn_1', doc_id='None', filing_id='None'}}>",
        ),
        (
            retrieval.Definition(doc_id="doc_id_1"),
            "<refinitiv.data.content.filings.retrieval.Definition object at {0} {{filename='None', dcn='None', doc_id='doc_id_1', filing_id='None'}}>",
        ),
        (
            retrieval.Definition(filing_id="filing_id_1"),
            "<refinitiv.data.content.filings.retrieval.Definition object at {0} {{filename='None', dcn='None', doc_id='None', filing_id='filing_id_1'}}>",
        ),
    ],
)
def test_retrieval_definition_repr(filings_definition, expected_repr):
    # given
    obj_id = hex(id(filings_definition))

    # when
    s = repr(filings_definition)

    # then
    assert s == expected_repr.format(obj_id)



import asyncio
import collections
import json
import os
import sys
from datetime import datetime, timedelta
from itertools import product
from urllib.parse import urlparse, parse_qs

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_datetime64tz_dtype, is_datetime64_any_dtype

from refinitiv.data import _configure as configure

PROJECT_CONFIG_PATH = os.path.join(os.getcwd(), configure._default_config_file_name)


def make_write(remove, config_path):
    def write(config, filepath=None, is_remove=True, silent=False):
        is_remove and remove(filepath)
        path = filepath or config_path
        if not is_remove:
            with open(path, "r") as f:
                d = {}
                try:
                    d = json.loads(f.read())
                except Exception as e:
                    if not silent:
                        raise e

            config.update(d)
        with open(path, "w") as f:
            s = json.dumps(config)
            f.write(s)

    return write


def make_remove(config_path):
    def remove(filepath=None):
        path = filepath or config_path
        if os.path.exists(path):
            try:
                os.remove(path)
            except PermissionError:
                pass

    return remove


def create_obj_with_value_by_path(path, value):
    res = root = {}

    if isinstance(path, list):
        *keys, last = path
    elif isinstance(path, str):
        *keys, last = path.split(".")
    else:
        raise Exception(path)

    for key in keys:
        next = {}
        root[key] = next
        root = next
    root[last] = value
    return res


remove_prj_config = make_remove(PROJECT_CONFIG_PATH)
write_prj_config = make_write(remove_prj_config, PROJECT_CONFIG_PATH)


def check_response_status(
    response,
    expected_status_code,
    expected_http_reason=None,
    expected_error_message=None,
    expected_error_code=None,
):
    """
    Parameters
    ----------
    response
    expected_status_code
    expected_http_reason
    expected_error_message
    expected_error_code
    -------

    """
    status_list = response.http_status
    if isinstance(status_list, dict):
        status_list = [status_list]
    for status in status_list:
        if isinstance(status, dict):
            check_status_code_and_reason(
                status, expected_status_code, expected_http_reason
            )
        elif isinstance(status, list):
            for status_item in status:
                check_status_code_and_reason(
                    status_item, expected_status_code, expected_http_reason
                )

    if expected_error_message is not None:
        assert len(response.errors) != 0, f"Error not found"
        for error in response.errors:
            assert (
                error.message == expected_error_message
            ), f"Actual error message {error.message}"
            assert (
                error.code == expected_error_code
            ), f"{error.code} != {expected_error_code}"


def check_status_code_and_reason(status, expected_status_code, expected_http_reason):
    assert (
        status["http_status_code"] == expected_status_code
    ), f"{status['http_status_code']} != {expected_status_code}"
    assert (
        status["http_reason"] == expected_http_reason
    ), f"Actual http reason {status['http_reason']}"


def check_non_empty_response_data(response):
    assert response.data.raw, f"Empty response.data.raw received"
    assert not response.data.df.empty, f"Empty response.data.df received"


def check_empty_response_data(response):
    assert (
        not response.data.raw
    ), f"Non-empty response.data.raw received: {response.data.raw}"
    assert (
        not response.data.df
    ), f"Non-empty response.data.df received: {response.data.df}"


def check_empty_dataframe_in_response(response):
    assert (
        response.data.df.empty
    ), f"Non-empty response.data.df received: {response.data.df}"


def check_extended_params_were_sent_in_request(response, expected_extended_params=None):
    if expected_extended_params:
        if isinstance(response.request_message, list):
            request = response.request_message[0][0]
        else:
            request = response.request_message

        body = json.loads(request.content.decode("utf-8")) if request.content else ""

        if request.method == "POST":
            if isinstance(body, list):
                body = body[0]
            for key in expected_extended_params.keys():
                assert (
                    body[key] == expected_extended_params[key]
                ), f"Body does not contain extended params {expected_extended_params[key]}, actual body:{body[key]}"
        else:
            parse_result = urlparse(str(request.url))
            query_params = parse_qs(parse_result.query)

            for key in expected_extended_params.keys():
                actual_params = query_params[key][0]
                if isinstance(expected_extended_params[key], int):
                    actual_params = int(actual_params)
                assert (
                    actual_params == expected_extended_params[key]
                ), f"Query does not contain extended params {expected_extended_params[key]}, actual query:{actual_params[0]}"

            assert len(body) == 0, f"GET request contain body - {body}"


async def get_async_response_from_definition(definition_object):
    response = await definition_object.get_data_async()
    return response


async def get_async_response_from_definitions(
    definition_object_first, definition_object_second
):
    result = await asyncio.gather(
        definition_object_first.get_data_async(),
        definition_object_second.get_data_async(),
    )

    return result


def check_response_dataframe_contains_columns_names(response, expected_columns_names):
    if hasattr(response, "data"):
        dataframe = response.data.df
    else:
        dataframe = response.df
    retrieved_column_names = dataframe.columns.array
    retrieved_column_names_lowercase = list(
        map(lambda x: x.lower(), retrieved_column_names)
    )
    for column_name in expected_columns_names:
        assert column_name.lower() in retrieved_column_names_lowercase, (
            f"Expected column name '{column_name}' is not found in list of columns retrieved from dataframe: "
            f"{retrieved_column_names} "
        )


def check_response_dataframe_contains_rows_names(response, expected_rows_names):
    dataframe = response.data.df
    retrieved_rows_names = dataframe.index.array
    for row_name in expected_rows_names:
        assert (
            row_name in retrieved_rows_names
        ), f"Expected row name {row_name} is not found in list of rows retrieved from dataframe: {retrieved_rows_names}"


def check_response_dataframe_universe_date_as_index(response, expected_universe):
    if isinstance(expected_universe, list):
        column_names = response.data.df.columns
        actual_universe = {universe for universe, _ in column_names}
        assert actual_universe == set(expected_universe), (
            f"Received columns list is different from expected: "
            f"{list(actual_universe)} != {list(expected_universe)}"
        )
    else:
        assert expected_universe == response.data.df.columns.names[0]


def check_response_dataframe_columns_date_as_index(
    response, requested_universe, requested_fields
):
    if isinstance(requested_fields, str):
        requested_fields = [requested_fields]
    if isinstance(requested_universe, list):
        expected_columns_names = list(product(requested_universe, requested_fields))
    else:
        expected_columns_names = requested_fields
    expected_columns_names_lowercase = column_names_to_lowercase(expected_columns_names)
    actual_columns = response.data.df.columns
    actual_columns_names_lowercase = column_names_to_lowercase(actual_columns)

    assert set(actual_columns_names_lowercase) == set(
        expected_columns_names_lowercase
    ), (
        f"Received columns list is different from expected: "
        f"{list(actual_columns_names_lowercase)} != {list(expected_columns_names_lowercase)}"
    )


def column_names_to_lowercase(column_names):
    if isinstance(column_names[0], str):
        return map(lambda col_name: col_name.lower(), column_names)
    elif isinstance(column_names[0], tuple):
        return map(
            lambda col_name: (col_name[0].lower(), col_name[1].lower()),
            column_names,
        )
    else:
        raise TypeError(
            "column_names must contain either strings, or tuples(string, string)"
        )


def check_index_column_contains_dates(response):
    dataframe = response
    if not isinstance(response, pd.DataFrame):
        dataframe = response.data.df
    index_values = list(dataframe.index.values)
    for date in index_values:
        assert date is not pd.NaT, f"Index column contains {pd.NaT}"
        assert is_datetime64_any_dtype(date)
        assert not is_datetime64tz_dtype(date)
        # assert isinstance(date, np.datetime64)

    headers = dataframe.columns
    if dataframe.columns.nlevels == 2:
        headers = dataframe.columns.levels[1]
    assert not any(
        elem in headers for elem in ("Date", "DATE", "date")
    ), f"Duplicated column with date data in dataframe columns"


def compare_list(received_list, expected_list):
    assert collections.Counter(received_list) == collections.Counter(
        expected_list
    ), f"Received list {received_list} != expected list {expected_list}"


def check_if_dataframe_is_not_none(dataframe):
    assert not dataframe.isnull().values.all(), f"DataFrame is NaN/None"


def display_event(event, stream, event_type, triggered_events=None):
    current_time = datetime.now().time()
    if triggered_events is not None:
        triggered_events.append({event_type: event})
    print(
        ">>> Stream ID {} - {} - {} event {} received at {}".format(
            stream.id, stream.name, event_type, event, current_time
        ),
    )


def add_callbacks(stream, triggered_events=None):
    stream.on_refresh(
        lambda event, stream: display_event(event, stream, "Refresh", triggered_events)
    )
    stream.on_update(
        lambda event, stream: display_event(event, stream, "Update", triggered_events)
    )
    stream.on_status(
        lambda event, stream: display_event(event, stream, "Status", triggered_events)
    )
    stream.on_error(
        lambda event, stream: display_event(event, stream, "Error", triggered_events)
    )
    stream.on_complete(
        lambda event, stream: display_event(event, stream, "Complete", triggered_events)
    )
    return stream


def assert_error(actual_error, argument_name):
    expected_message = (
        f"__init__() missing 1 required positional argument: '{argument_name}'"
    )
    if sys.version_info >= (3, 10):
        expected_message = f"Definition.__init__() missing 1 required positional argument: '{argument_name}'"

    testing_message = str(actual_error.value)
    assert testing_message == expected_message, actual_error


def is_expected_param_in_request_url(response, expected_params):
    requests = response.request_message
    if not isinstance(response.request_message, list):
        requests = [requests]
    for universe_request in requests:
        for request_item in universe_request:
            request_params = request_item.url.params._dict
            for param in expected_params:
                assert (
                    param in request_params.keys()
                ), f"There is no '{param}' param is request url"


def convert_to_datetime(date):
    if isinstance(date, timedelta):
        converted_date = datetime.utcnow() + date
    else:
        converted_date = pd.to_datetime(date)

    return converted_date


def check_response_data_start_end_date(response, start_date, end_date, request=None):
    if not isinstance(response, pd.DataFrame):
        data_datetimes = response.data.df.index
    else:
        data_datetimes = response.index

    start_datetime = convert_to_datetime(start_date)
    end_datetime = convert_to_datetime(end_date)

    if not request.cls.pytestmark[0].args[0] == "datagrid":

        assert (
            data_datetimes[0] >= start_datetime
        ), f"Data starts {start_datetime} earlier than requested: {data_datetimes[0]}"

    assert (
        data_datetimes[-1] <= end_datetime
    ), f"Data ends later than requested: {data_datetimes[-1]}"


def is_universe_empty(dataframe, universe):
    return dataframe[universe].isnull().values.all()


def check_response_is_success(response):
    assert response.is_success


def check_response_data_universe(response, expected_universe):
    actual_universe = {
        data_object["universe"]["ric"] for data_object in response.data.raw
    }
    assert set(expected_universe) == actual_universe


def check_response_values(
    response,
    expected_universe_ric=None,
    expected_interval=None,
    expected_summary_timestamp_label=None,
):
    """

    Parameters
    ----------
    response
    expected_universe_ric
    expected_interval
    expected_summary_timestamp_label
    -------

    """
    data = response.data.raw
    try:
        if expected_universe_ric is not None:
            check_universe_in_response(response, expected_universe_ric)

        if expected_interval is not None:
            assert (
                data["interval"] == expected_interval
            ), f"{data['interval']} != {expected_interval}"

        if expected_summary_timestamp_label is not None:
            assert (
                data["summaryTimestampLabel"] == expected_summary_timestamp_label
            ), f"{data['summaryTimestampLabel']} != {expected_summary_timestamp_label}"
    except KeyError as err:
        raise AssertionError(
            f"The data returned does not have expected property: {data}."
            f" \n The error appeared: {str(err)}"
        )


def check_universe_in_response(response, expected_universe_ric):
    data: dict = response.data.raw
    if isinstance(expected_universe_ric, str):
        assert (
            data["universe"]["ric"] == expected_universe_ric
        ), f"{data['universe']['ric']} != {expected_universe_ric}"

    elif isinstance(expected_universe_ric, list):
        retrieved_universes = list(map(lambda item: item["universe"]["ric"], data))
        assert len(retrieved_universes) == len(expected_universe_ric)
        compare_list(retrieved_universes, expected_universe_ric)
    else:
        raise Exception(
            f"Unexpected type of the expected_universe_ric was provided. "
            f"{type(expected_universe_ric)}, value: {expected_universe_ric}"
        )


def check_dataframe_column_date_for_datetime_type(response):
    if hasattr(response, "data"):
        dataframe = response.data.df
    elif hasattr(response, "df"):
        dataframe = response.df
    else:
        dataframe = response
    actual_column_names = dataframe.columns.array
    for column in actual_column_names:
        if any(
            [
                i
                for i in ["Date", "date", "_DT", "DATE"]
                if i in column
            ]
        ) and all([i if not i in column else False for i in ["DateType", "Dates"]]):

            assert is_datetime64_any_dtype(
                dataframe[column]
            ), f"Column {column} has {type(dataframe[column])} dtype"
            assert not is_datetime64tz_dtype(dataframe[column])


def check_the_number_of_items_in_dataframe(universe, list_of_instruments, count):
    dict_count_of_instruments = {
        item: list_of_instruments.count(item) for item in list_of_instruments
    }
    assert (
        dict_count_of_instruments.get(universe) == count
    ), dict_count_of_instruments.get(universe)


def check_universe_order_in_df(response, expected_universe):
    dataframe = response
    if not isinstance(response, pd.DataFrame):
        dataframe = response.data.df
    if isinstance(expected_universe, list):
        actual_universes = []
        for universe, _ in dataframe.columns:
            if universe not in actual_universes:
                actual_universes.append(universe)
    else:
        actual_universes = dataframe.columns.name
    assert actual_universes == expected_universe


def check_response_value(response, column_name, expected_value):
    if isinstance(expected_value, str):
        expected_value = [expected_value]
    actual_value = []
    for values in response[column_name]:
        actual_value.append(values)

    compare_list(list(set(actual_value)), expected_value)


def check_contrib_response(request, response, event_list, message):
    assert isinstance(event_list[0], dict)
    assert event_list[0]["Text"] == message, f"Actual message: {event_list[0]['Text']}"

    if "negative_case" in request.node.callspec.id:
        assert response.error == message, f"Actual error: {event_list[0]['Text']}"
    else:
        assert (
            response.nak_message == message
        ), f"Actual nak_message: {event_list[0]['Text']}"


def check_event_code(event_list, expected_events):
    event_codes = [event[0] for event in event_list]
    expected_codes = [event[0] for event in expected_events]

    compare_list(event_codes, expected_codes)


def check_event_message(event_list, expected_events):
    event_messages = []
    for event, msg in event_list:
        if isinstance(msg, dict):
            event_messages.append(msg.get("api_cfg"))
        else:
            event_messages.append(msg)

    expected_messages = []
    for event, msg in expected_events:
        if isinstance(msg, dict):
            expected_messages.append(msg.get("api_cfg"))
        else:
            expected_messages.append(msg)

    compare_list(event_messages, expected_messages)



import pytest

from refinitiv.data.content.pricing.chain._chain_record import can_create_chain_record


@pytest.mark.parametrize(
    "input_data,expected_value",
    [
        ({}, False),
        ({"REF_COUNT": 10}, False),
        (
            {"REF_COUNT": 10, **{f"LONGLINK{number}": None for number in range(1, 15)}},
            False,
        ),
        (
            {
                "REF_COUNT": 10,
                **{f"LONGLINK{number}": None for number in range(1, 15)},
                "LONGPREVLR": None,
                "LONGNEXTLR": None,
            },
            True,
        ),
    ],
)
def test_method_is_valid_chain_record(input_data, expected_value):

    try:
        result = can_create_chain_record(input_data)
    except Exception as e:
        assert False, e

    assert result is expected_value



from datetime import datetime
from unittest.mock import patch, MagicMock

import pandas as pd
import pytest

from refinitiv.data._fin_coder_layer._ohlc_builder import (
    merge_dataframes,
    create_df,
    replace_values_by_nan,
    Ticks_OHLCBuilder,
    OHLCBuilder,
)
from refinitiv.data._fin_coder_layer._pricing_recorder import PricingRecorder
from refinitiv.data._fin_coder_layer._stream_update_handler import (
    BuildDF_StreamUpdateHandler,
    CollectUpdates_StreamUpdateHandler,
)
from refinitiv.data._content_type import ContentType
from refinitiv.data.content._universe_streams import _UniverseStreams
from refinitiv.data.session import set_default
from tests.unit.conftest import StubSession

MOCKED_RECORDED_UPDATES = {
    "GBP=": [
        {
            "Type": "Update",
            "Key": {"Service": "ELEKTRON_DD", "Name": "GBP="},
            "Fields": {"BID": 1.3317, "ASK": 1.3318},
            "Timestamp": datetime(2021, 12, 16, 18, 28, 33, 109421),
        },
        {
            "Type": "Update",
            "Key": {"Service": "ELEKTRON_DD", "Name": "GBP="},
            "Fields": {"BID": 1.3317, "ASK": 1.3318},
            "Timestamp": datetime(2021, 12, 16, 18, 28, 34, 412376),
        },
    ],
    "EUR=": [
        {
            "Type": "Update",
            "Key": {"Service": "ELEKTRON_DD", "Name": "EUR="},
            "Fields": {"BID": 1.1305, "ASK": 1.1309},
            "Timestamp": datetime(2021, 12, 16, 18, 28, 34, 412376),
        },
        {
            "Type": "Update",
            "Key": {"Service": "ELEKTRON_DD", "Name": "EUR="},
            "Fields": {"BID": 1.1307, "ASK": 1.1308, "VALUE_TS3": "2010-01-01"},
            "Timestamp": datetime(2021, 12, 16, 18, 28, 34, 428335),
        },
    ],
}


@pytest.fixture(
    scope="function",
)
def recorder(request):
    session = StubSession(is_open=True)
    set_default(session)

    stream = _UniverseStreams(
        content_type=ContentType.NONE,
        session=session,
        universe=["EUR=", "GBP="],
        fields=["BID", "ASK", "OPEN_PRC"],
    )
    recorder = PricingRecorder(stream)
    yield recorder
    set_default(None)


@patch(
    "refinitiv.data.delivery._stream.StreamStateManager.is_closed",
    False,
)
def test_record_ohlc_data(recorder):
    df = pd.DataFrame()
    recorder.record(frequency="tick", ticks_per_bar="2")

    assert recorder._ohlc_builder.ohlc_df is None
    recorder._ohlc_builder.save_ohlc_data(df)
    assert recorder._ohlc_builder.ohlc_df is df


@patch("refinitiv.data.delivery._stream.StreamStateManager.is_closed", False)
@patch(
    "refinitiv.data._fin_coder_layer._pricing_recorder.NoBlocking_RecordingControl",
    return_value=MagicMock(),
)
def test_record_with_ticks(recording_control, recorder):
    recorder.record("tick")

    assert isinstance(recorder._update_handler, CollectUpdates_StreamUpdateHandler)
    recording_control.assert_called_once_with(recorder._update_handler)
    recorder._recording_control.start_recording.assert_called_once()


@patch("refinitiv.data.delivery._stream.StreamStateManager.is_closed", False)
@patch(
    "refinitiv.data._fin_coder_layer._pricing_recorder.NoBlocking_RecordingControl",
    return_value=MagicMock(),
)
def test_record_with_ticks_and_ticks_per_bar(recording_control, recorder):
    recorder.record(frequency="tick", ticks_per_bar="10")

    assert isinstance(recorder._ohlc_builder, Ticks_OHLCBuilder)
    assert isinstance(recorder._update_handler, BuildDF_StreamUpdateHandler)
    recording_control.assert_called_once_with(recorder._update_handler)
    recorder._recording_control.start_recording.assert_called_once()


@patch("refinitiv.data.delivery._stream.StreamStateManager.is_closed", False)
@patch(
    "refinitiv.data._fin_coder_layer._pricing_recorder.NoBlocking_RecordingControl",
    return_value=MagicMock(),
)
def test_record_with_ticks_and_ticks_per_bar(recording_control, recorder):
    recorder.record(frequency="tick", ticks_per_bar="10")

    assert isinstance(recorder._ohlc_builder, Ticks_OHLCBuilder)
    assert isinstance(recorder._update_handler, BuildDF_StreamUpdateHandler)
    recording_control.assert_called_once_with(recorder._update_handler)
    recorder._recording_control.start_recording.assert_called_once()


@patch("refinitiv.data.delivery._stream.StreamStateManager.is_closed", False)
@patch(
    "refinitiv.data._fin_coder_layer._pricing_recorder.Blocking_RecordingControl",
    return_value=MagicMock(),
)
def test_record_with_ticks_and_duration(recording_control, recorder):
    recorder.record(frequency="tick", duration="10s")

    assert isinstance(recorder._update_handler, CollectUpdates_StreamUpdateHandler)
    recording_control.assert_called_once_with(recorder._update_handler)
    recorder._recording_control.start_recording.assert_called_once_with(10)


@patch("refinitiv.data.delivery._stream.StreamStateManager.is_closed", False)
@patch(
    "refinitiv.data._fin_coder_layer._pricing_recorder.Blocking_RecordingControl",
    return_value=MagicMock(),
)
def test_record_with_ticks_and_duration_and_ticks_per_bar(recording_control, recorder):
    recorder.record(frequency="tick", ticks_per_bar="10", duration="15s")

    assert isinstance(recorder._update_handler, BuildDF_StreamUpdateHandler)
    recording_control.assert_called_once_with(recorder._update_handler)
    recorder._recording_control.start_recording.assert_called_once()


@patch("refinitiv.data.delivery._stream.StreamStateManager.is_closed", False)
@patch(
    "refinitiv.data._fin_coder_layer._pricing_recorder.RepeatBlocking_RecordingControl",
    return_value=MagicMock(),
)
def test_record_with_frequency_and_duration(recording_control, recorder):
    recorder.record(frequency="5s", duration="10s")

    assert isinstance(recorder._ohlc_builder, OHLCBuilder)
    assert isinstance(recorder._update_handler, CollectUpdates_StreamUpdateHandler)
    recording_control.assert_called_once_with(
        10,
        None,
        recorder._update_handler,
        recorder._ohlc_builder,
        recorder._logger,
        recorder,
    )
    recorder._recording_control.start_recording.assert_called_once()


@patch("refinitiv.data.delivery._stream.StreamStateManager.is_closed", False)
@patch(
    "refinitiv.data._fin_coder_layer._pricing_recorder.RepeatNonBlocking_RecordingControl",
    return_value=MagicMock(),
)
def test_record_with_frequency(recording_control, recorder):
    recorder.record(frequency="5s")

    assert isinstance(recorder._ohlc_builder, OHLCBuilder)
    assert isinstance(recorder._update_handler, CollectUpdates_StreamUpdateHandler)
    recording_control.assert_called_once_with(
        None,
        recorder._update_handler,
        recorder._ohlc_builder,
        recorder._logger,
        recorder,
    )
    recorder._recording_control.start_recording.assert_called_once()


@patch("refinitiv.data.delivery._stream.StreamStateManager.is_closed", False)
@patch(
    "refinitiv.data._fin_coder_layer._pricing_recorder.NoBlocking_RecordingControl",
    return_value=MagicMock(),
)
def test_stop(recording_control, recorder):
    recorder.record(frequency="tick", ticks_per_bar="10")

    recorder.stop()
    recording_control.assert_called_once()

    recorder._recording_control.stop_recording.assert_called_once()


@pytest.mark.parametrize(
    "input, parsed_value",
    [
        ("20s", 20),
        ("2m", 120),
        ("1h", 3600),
    ],
)
def test_parse_input_frequency_and_duration(recorder, input, parsed_value):
    result = recorder._parse_input_frequency_and_duration("20s")
    assert result == 20

    result = recorder._parse_input_frequency_and_duration("2min")
    assert result == 120

    result = recorder._parse_input_frequency_and_duration("1h")
    assert result == 3600


def test_parse_input_frequency_and_duration_error(recorder):
    with pytest.raises(
        ValueError,
        match="Please provide 'duration' or 'frequency' value "
        "in valid format. For example: '10s', '2min', '1h'",
    ):
        recorder._parse_input_frequency_and_duration("invalid data")


def test_merge_dataframes(recorder):
    df = merge_dataframes([pd.DataFrame(), pd.DataFrame()])

    assert "ohlc" in dir(df)
    assert df.index.name == "Timestamp"


@patch(
    "refinitiv.data.delivery._stream.StreamStateManager.is_closed",
    False,
)
def test_create_df(recorder):
    mocked_data = [[1.3319, 1.332], [1.3317, 1.3321]]
    mocked_fields = ["BID", "ASK"]
    mocked_timestamps = [
        datetime(2021, 12, 16, 18, 9, 41, 56778),
        datetime(2021, 12, 16, 18, 9, 35, 585482),
    ]
    mocked_stream_name = "EUR="
    recorder.record(frequency="tick", ticks_per_bar="2")

    df = create_df(mocked_data, mocked_timestamps, mocked_fields, mocked_stream_name)
    assert not df.empty

    df = create_df([], [], [], mocked_stream_name)
    assert df.empty


@patch(
    "refinitiv.data.delivery._stream.StreamStateManager.is_closed",
    False,
)
def test_on_update_handler(recorder):
    mocked_update = MagicMock()
    recorder.record(frequency="tick", ticks_per_bar="5")

    streaming_price_instance = recorder._stream._stream_by_name["EUR="]
    recorder._update_handler._on_update_handler(streaming_price_instance, mocked_update)

    assert mocked_update in recorder._update_handler.updates_by_stream_name["EUR="]

    recorder._update_handler._on_update_handler(streaming_price_instance, mocked_update)

    assert len(recorder._update_handler.updates_by_stream_name["EUR="]) == 2


@patch(
    "refinitiv.data.delivery._stream.StreamStateManager.is_closed",
    False,
)
def test_on_update_handler_with_ticks(recorder):
    streaming_price_eur = recorder._stream._stream_by_name["EUR="]
    streaming_price_gbp = recorder._stream._stream_by_name["GBP="]
    recorder.record(frequency="tick", ticks_per_bar="3")

    assert recorder._update_handler.counter == 0

    recorder._update_handler._on_update_handler(
        stream=streaming_price_eur, message={"Fields": {"BID": 1.3559, "ASK": 1.3561}}
    )

    assert recorder._update_handler.counter == 1

    recorder._update_handler._on_update_handler(
        streaming_price_gbp, {"Fields": {"BID": 1.3559, "ASK": 1.3561}}
    )
    assert recorder._update_handler.counter == 2


def test_validate_arguments_invalid_frequency(recorder):
    with pytest.raises(
        ValueError,
        match="Please provide 'tick' value as frequency when you are using 'ticks_per_bar' argument.",
    ):
        recorder._validate_arguments(
            frequency="2min", duration="1min", ticks_per_bar="100"
        )


def test_validate_arguments_(recorder):
    with pytest.raises(
        ValueError,
        match="Please check your arguments, 'duration' should be higher that 'frequency'.",
    ):
        recorder._validate_arguments(
            frequency="2min", duration="1min", ticks_per_bar="1"
        )


@patch(
    "refinitiv.data.delivery._stream.StreamStateManager.is_closed",
    False,
)
def test_get_history(recorder):
    recorder.record(frequency="tick", ticks_per_bar="1")

    recorder._update_handler.updates_by_stream_name = MOCKED_RECORDED_UPDATES
    df = recorder.get_history()

    assert not df.empty
    assert "ohlc" in dir(df)
    assert df.index.name == "Timestamp"


@patch(
    "refinitiv.data.delivery._stream.StreamStateManager.is_closed",
    False,
)
def test_get_ohlc_history_no_data(recorder):
    recorder.record(frequency="tick", ticks_per_bar="1")

    df = recorder.get_history()
    assert df.empty


@patch(
    "refinitiv.data.delivery._stream.StreamStateManager.is_closed",
    False,
)
def test_check_recorded_ohlc_df(recorder):
    recorder.record(frequency="tick", ticks_per_bar="3")

    recorder._ohlc_builder.ohlc_df = pd.DataFrame()

    df = recorder._check_df(recorder._ohlc_builder.ohlc_df)
    assert df is recorder._ohlc_builder.ohlc_df


def test_validate_count_argument_invalid_value(recorder):
    with pytest.raises(
        ValueError,
        match="Invalid argument. Please provide 'ticks_per_bar'"
        " in the following format: '10', '100', '500'",
    ):
        recorder._validate_count_argument("invalid value")


def test_validate_count_argument_invalid_count(recorder):
    with pytest.raises(
        ValueError, match="Invalid argument. 'ticks_per_bar' should be more then 0"
    ):
        recorder._validate_count_argument("0")


@patch(
    "refinitiv.data.delivery._stream.StreamStateManager.is_closed",
    False,
)
def test_record_with_invalid_freqency_and_valid_count(recorder):
    with pytest.raises(
        ValueError,
        match="Please provide 'tick' value as frequency when you are using 'ticks_per_bar' argument.",
    ):
        recorder.record(frequency="3s", ticks_per_bar="100")


def test_replace_values_by_nan(recorder):
    data = [[17.0, "18.9", "2010-01-01"]]
    replace_values_by_nan(data)
    assert data == [[17.0, "18.9", None]]



import abc
from collections import Counter
from itertools import groupby
from typing import Optional, Dict, List

from pandas import DataFrame

from ._context import Context
from .._containers import UniverseContainer
from .._history_provider import group_universe, chunks
from .._intervals_consts import NON_INTRA_DAY_INTERVALS
from ..._tools import ohlc
from ..._tools._dataframe import convert_dtypes


class ADCContext(Context, abc.ABC):
    @property
    def can_build_df(self) -> bool:
        return bool(self._adc_data and not (self._hp_data or self._cust_inst_data))

    @property
    def can_get_data(self) -> bool:
        if not self.fields.adc:
            return False

        return bool(self.universe.adc) and not (
            self.universe.is_universe_expander
            and self.fields.hp
            and self.fields.is_disjoint_adc
        )

    @property
    def raw(self) -> Optional[Dict]:
        return self._adc_data and self._adc_data.raw

    @property
    def df(self) -> Optional[DataFrame]:
        if type(self._adc_data.df) is DataFrame and not self._adc_data.df.empty:
            return convert_dtypes(self._adc_data.df)
        return self._adc_data.df

    @abc.abstractmethod
    def is_raw_headers_in_fields(self, raw, fields) -> bool:
        # for override
        pass

    def build_df(self, *args) -> DataFrame:
        df = self.df
        if not self.is_raw_headers_in_fields(self.raw, self.fields):
            df = DataFrame()

        df.sort_index(ascending=True, inplace=True)
        df = convert_dtypes(df)
        df.ohlc = ohlc.__get__(df, None)
        return df

    @abc.abstractmethod
    def get_cols(self, headers, fields, *args) -> List[str]:
        # for override
        pass

    def prepare_data(self, interval, *args):
        """
        Transform adc data to further merging.

        Transformed data looks like:
        {
            "RIC": [{"Date": "value", "field": "value", "another_field": "value"}],
            "NEXT_RIC": [{"Date": "value", "field": "value", "next_field": "value"}]
        }
        """
        headers = self.dfbuilder.get_headers(self.raw)
        cols = ["Type", *self.get_cols(headers, self.fields.adc)]
        data = []
        universe = []
        index_universe = 0
        universe_count = group_universe(self.universe.adc)
        len_universe_count = len(universe_count)
        counter_items = Counter()

        if interval in NON_INTRA_DAY_INTERVALS:

            def process_item_date(s: str) -> str:
                return s.split("T")[0]

        else:

            def process_item_date(s: str) -> str:
                return s.replace("T", " ").replace("Z", "")

        for ric, items in groupby(self.raw["data"], lambda i: i[0]):
            items = list(items)
            len_items = len(items)
            counter_items.update([len_items])
            result_items = []

            if index_universe < len_universe_count:
                name_universe, count = universe_count[index_universe]

            else:
                count = 0
                name_universe = ""

            for item in items:
                item[0] = "adc"
                item = {k: v for k, v in zip(cols, item)}
                item_date = item.get(self.date_name)

                if item_date:
                    item[self.date_name] = process_item_date(item_date)

                result_items.append(item)

            repeat_universe = 1
            if ric == name_universe:

                if count == len_items:
                    repeat_universe = count

                    for i in result_items:
                        data.append([i])

                elif (
                    count < len_items
                    and counter_items.most_common(1)[0][0] * (count + 1) == len_items
                ):
                    repeat_universe = count + 1
                    for chunk in chunks(result_items, len_items // repeat_universe):
                        data.append(chunk)

                elif count < len_items:
                    repeat_universe = count
                    for chunk in chunks(result_items, len_items // count):
                        data.append(chunk)

                elif count > len_items:
                    for chunk in chunks(result_items, len_items):
                        data.append(chunk)

                index_universe += 1

            else:
                data.append(result_items)

            universe.extend([ric] * repeat_universe)

        return data, universe

    def prepare_to_build(self, interval, *args):
        data, universe = self.prepare_data(interval)
        self.universe = UniverseContainer(universe)
        self._data = data
        self._headers = self.prepare_headers(self.raw)



from refinitiv.data.delivery import endpoint_request

definition = endpoint_request.Definition(
    url="/data/estimates/v1/view-actuals-kpi/annual",
    method=endpoint_request.RequestMethod.GET,
    query_parameters={"universe": "IBM"},
)



__all__ = (
    "Definition",
    "contribute",
    "contribute_async",
    "ContribType",
    "ContribResponse",
)

from ._stream.contrib import contribute, contribute_async, ContribResponse, ContribType
from ._stream.omm_stream_definition import Definition



# coding: utf8

from ._surface_request_item import SurfaceRequestItem

from ._enums import UnderlyingType
from ._fx_surface_parameters import FxSurfaceParameters as FxCalculationParams
from ._fx_surface_definition import FxVolatilitySurfaceDefinition as FxSurfaceDefinition


class FxSurfaceRequestItem(SurfaceRequestItem):
    def __init__(
        self,
        surface_layout=None,
        surface_parameters=None,
        underlying_definition=None,
        surface_tag=None,
    ):
        super().__init__(
            surface_layout=surface_layout,
            surface_tag=surface_tag,
            underlying_type=UnderlyingType.FX,
        )
        self.surface_parameters = surface_parameters
        self.underlying_definition = underlying_definition

    @property
    def surface_parameters(self):
        """
        The section that contains the properties that define how the volatility surface is generated
        :return: object FxCalculationParams
        """
        return self._get_object_parameter(FxCalculationParams, "surfaceParameters")

    @surface_parameters.setter
    def surface_parameters(self, value):
        self._set_object_parameter(FxCalculationParams, "surfaceParameters", value)

    @property
    def underlying_definition(self):
        """
        The section that contains the properties that define the underlying instrument
        :return: object FxSurfaceDefinition
        """
        return self._get_object_parameter(FxSurfaceDefinition, "underlyingDefinition")

    @underlying_definition.setter
    def underlying_definition(self, value):
        self._set_object_parameter(FxSurfaceDefinition, "underlyingDefinition", value)



from refinitiv.data.content import symbol_conversion


def symbol_conversion_definition():
    definition = symbol_conversion.Definition(
        symbols=["60000008", "60003513"],
        from_symbol_type=symbol_conversion.SymbolTypes.LIPPER_ID,
        preferred_country_code=symbol_conversion.CountryCode.USA,
    )
    return definition


def invalid_symbol_conversion_definition():
    definition = symbol_conversion.Definition(
        symbols="invalid",
        from_symbol_type=symbol_conversion.SymbolTypes.LIPPER_ID,
        to_symbol_types=[symbol_conversion.SymbolTypes.CUSIP],
        preferred_country_code="UK",
    )
    return definition



__all__ = (
    "Axis",
    "BidAskMid",
    "DayWeight",
    "Definition",
    "Format",
    "FxCalculationParams",
    "FxSurfaceDefinition",
    "FxSwapCalculationMethod",
    "FxVolatilityModel",
    "InterpolationWeight",
    "PriceSide",
    "SurfaceLayout",
    "TimeStamp",
)

from ._definition import Definition
from ._enums import (
    FxVolatilityModel,
    FxSwapCalculationMethod,
    PriceSide,
    TimeStamp,
    Axis,
    Format,
)
from ._models import BidAskMid, InterpolationWeight, DayWeight, SurfaceLayout
from ..._surfaces._fx_surface_definition import (
    FxVolatilitySurfaceDefinition as FxSurfaceDefinition,
)
from ..._surfaces._fx_surface_parameters import (
    FxSurfaceParameters as FxCalculationParams,
)



import hashlib
import warnings

from refinitiv.data._external_libraries.python_configuration import config_from_dict

__all__ = ("config",)

data = {
    "config-change-notifications-enabled": False,
    "raise_exception_on_error": True,
    "http": {
        "max-connections": 100,
        "max-keepalive-connections": 20,
        "request-timeout": 20,
        "auto-retry": {"number-of-retries": 6, "on-errors": [429], "backoff-factor": 1},
    },
    "logs": {
        "level": "info",
        "filter": "*",
        "transports": {
            "console": {"enabled": False, "colorize": False},
            "file": {
                "enabled": False,
                "name": "refinitiv-data-lib.log",
                "size": "10M",
                "interval": "1d",
                "maxFiles": 10,
            },
        },
    },
    "usage_logger": {
        "enabled": False,
    },
    "sessions": {
        "default": "desktop.workspace",
        "platform": {
            "default": {
                "signon_control": True,
                "auto-reconnect": True,
                "server-mode": False,
                "verify_scope": True,
                "base-url": "https://api.refinitiv.com",
                "auth": {
                    "url": "/auth/oauth2/v1",
                    "authorize": "/authorize",
                    "token": "/token",
                },
            }
        },
        "desktop": {
            "workspace": {
                "app-key": "DEFAULT_WORKSPACE_APP_KEY",
                "base-url": "http://localhost:9000",
                "platform-paths": {"rdp": "/api/rdp", "udf": "/api/udf"},
                "handshake-url": "/api/handshake",
            },
        },
    },
    "apis": {
        "file-store": {
            "url": "/file-store/v1",
            "endpoints": {
                "buckets": "/buckets",
                "file-sets": "/file-sets",
                "files": "/files",
                "packages": "/packages",
            },
        },
        "data": {
            "historical-pricing": {
                "url": "/data/historical-pricing/v1",
                "endpoints": {
                    "events": "/views/events",
                    "interday-summaries": "/views/interday-summaries",
                    "intraday-summaries": "/views/intraday-summaries",
                },
            },
            "pricing": {
                "url": "/data/pricing",
                "endpoints": {"snapshots": "/snapshots/v1/", "chains": "/chains/v1/"},
            },
            "quantitative-analytics-dates-and-calendars": {
                "url": "/data/quantitative-analytics-dates-and-calendars/v1",
                "endpoints": {
                    "add-periods": {"sync": "/add-periods"},
                    "holidays": {"sync": "/holidays"},
                    "count-periods": {"sync": "/count-periods"},
                    "date-schedule": {"sync": "/date-schedule"},
                    "is-working-day": {"sync": "/is-working-day"},
                },
            },
            "quantitative-analytics-financial-contracts": {
                "url": "/data/quantitative-analytics/v1",
                "endpoints": {
                    "financial-contracts": {
                        "sync": "/financial-contracts",
                        "async": "/async-financial-contracts",
                    },
                    "async-operation": "/async-operation",
                    "async-resource": "/async-resource",
                },
                "auto-retry": True,
            },
            "quantitative-analytics-curves-and-surfaces": {
                "url": "/data/quantitative-analytics-curves-and-surfaces/v1",
                "endpoints": {
                    "bond-curves": {
                        "curves": "/curves/bond-curves/curves",
                    },
                    "cross-currency-curves": {
                        "curves": "/curves/cross-currency-curves/curves",
                        "definitions": {
                            "create": "/curves/cross-currency-curves/definitions/create",
                            "delete": "/curves/cross-currency-curves/definitions/delete",
                            "get": "/curves/cross-currency-curves/definitions/get",
                            "search": "/curves/cross-currency-curves/definitions/search",
                            "update": "/curves/cross-currency-curves/definitions/update",
                        },
                        "triangulate-definitions": {
                            "search": "/curves/cross-currency-curves/triangulate-definitions/search"
                        },
                    },
                    "forward-curves": "/curves/forward-curves",
                    "surfaces": {"sync": "/surfaces", "async": "/async-surfaces"},
                    "zc-curves": "/curves/zc-curves",
                    "zc-curve-definitions": "/curves/zc-curve-definitions",
                    "async-operation": "/async-operation",
                    "async-resource": "/async-resource",
                },
                "auto-retry": True,
            },
            "news": {
                "url": "/data/news/v1",
                "underlying-platform": "udf",
                "endpoints": {
                    "headlines": "/headlines",
                    "stories": "/stories",
                    "top-news": "/top-news",
                    "images": "/images",
                },
            },
            "environmental-social-governance": {
                "url": "/data/environmental-social-governance/v1",
                "endpoints": {
                    "universe": "/universe",
                    "basic": "/views/basic",
                    "measures-full": "/views/measures-full",
                    "measures-standard": "/views/measures-standard",
                    "scores-full": "/views/scores-full",
                    "scores-standard": "/views/scores-standard",
                },
            },
            "datagrid": {
                "url": "/data/datagrid/beta1",
                "underlying-platform": "udf",
                "endpoints": {"standard": "/"},
            },
            "ownership": {
                "url": "/data/ownership/v1",
                "endpoints": {
                    "consolidated": {
                        "breakdown": "/views/consolidated/breakdown",
                        "concentration": "/views/consolidated/concentration",
                        "investors": "/views/consolidated/investors",
                        "recent-activity": "/views/consolidated/recent-activity",
                        "shareholders-history-report": "/views/consolidated/shareholders-history-report",
                        "shareholders-report": "/views/consolidated/shareholders-report",
                        "top-n-concentration": "/views/consolidated/top-n-concentration",
                    },
                    "fund": {
                        "breakdown": "/views/fund/breakdown",
                        "concentration": "/views/fund/concentration",
                        "investors": "/views/fund/investors",
                        "holdings": "/views/fund/holdings",
                        "recent-activity": "/views/fund/recent-activity",
                        "shareholders-history-report": "/views/fund/shareholders-history-report",
                        "shareholders-report": "/views/fund/shareholders-report",
                        "top-n-concentration": "/views/fund/top-n-concentration",
                    },
                    "insider": {
                        "shareholders-report": "/views/insider/shareholders-report",
                        "transaction-report": "/views/insider/transaction-report",
                    },
                    "investor": {"holdings": "/views/investor/holdings"},
                    "org-info": "/views/org-info",
                },
            },
            "estimates": {
                "url": "/data/estimates/v1",
                "endpoints": {
                    "view-actuals": {
                        "annual": "/view-actuals/annual",
                        "interim": "/view-actuals/interim",
                    },
                    "view-actuals-kpi": {
                        "annual": "/view-actuals-kpi/annual",
                        "interim": "/view-actuals-kpi/interim",
                    },
                    "view-summary": {
                        "annual": "/view-summary/annual",
                        "historical-snapshots-non-periodic-measures": "/view-summary/historical-snapshots-non-periodic-measures",
                        "historical-snapshots-periodic-measures-annual": "/view-summary/historical-snapshots-periodic-measures-annual",
                        "historical-snapshots-periodic-measures-interim": "/view-summary/historical-snapshots-periodic-measures-interim",
                        "historical-snapshots-recommendations": "/view-summary/historical-snapshots-recommendations",
                        "interim": "/view-summary/interim",
                        "non-periodic-measures": "/view-summary/non-periodic-measures",
                        "recommendations": "/view-summary/recommendations",
                    },
                    "view-summary-kpi": {
                        "annual": "/view-summary-kpi/annual",
                        "historical-snapshots-kpi": "/view-summary-kpi/historical-snapshots-kpi",
                        "interim": "/view-summary-kpi/interim",
                    },
                },
            },
            "custom-instruments": {
                "url": "/data/custom-instruments/v1",
                "endpoints": {
                    "instruments": "/instruments",
                    "search": "/search",
                    "events": "/events",
                    "interday-summaries": "/interday-summaries",
                    "intraday-summaries": "/intraday-summaries",
                },
            },
            "filings": {
                "headers": {
                    "ClientID": "API_Playground",
                    "X-Api-Key": "155d9dbf-f0ac-46d9-8b77-f7f6dcd238f8",
                    "Accept": "*/*",
                },
                "url": "/data/filings/v1",
                "endpoints": {"retrieval": "/retrieval"},
            },
            "qpl-functions": {
                "endpoints": {
                    "fx_swp_to_swp": "https://qps-userfunctions-use1.dev.refinitiv.com/api/fx/fx-curve-from-fx-swaps",
                },
            },
        },
        "discovery": {
            "search": {
                "url": "/discovery/search/v1/",
                "endpoints": {
                    "search": "/",
                    "lookup": "/lookup",
                    "metadata": "/metadata/views",
                },
            }
        },
        "streaming": {
            "pricing": {
                "url": "/streaming/pricing/v1",
                "endpoints": {
                    "main": {"path": "/", "protocols": ["OMM"], "locations": []}
                },
            },
            "custom-instruments": {
                "url": "/streaming/custom-instruments/v1",
                "endpoints": {
                    "resource": {
                        "path": "/resource",
                        "protocols": ["OMM"],
                        "locations": [],
                    }
                },
            },
            "trading-analytics": {
                "url": "/streaming/trading-analytics/trade-data/beta1",
                "endpoints": {
                    "redi": {
                        "path": "/redi",
                        "protocols": ["OMM", "RDP"],
                        "locations": [],
                    }
                },
            },
            "quantitative-analytics": {
                "url": "/streaming/quantitative-analytics/beta1",
                "endpoints": {
                    "financial-contracts": {
                        "path": "/financial-contracts",
                        "protocols": ["OMM", "RDP"],
                        "locations": [],
                    }
                },
            },
            "benchmark": {
                "url": "/streaming/benchmark/v1",
                "endpoints": {
                    "resource": {
                        "path": "/resource",
                        "protocols": ["RDP"],
                        "locations": [],
                    }
                },
            },
        },
        "data-store": {"url": "/data-store/v1", "endpoints": {"graphql": "/graphql"}},
    },
    "bulk": {
        "esg": {
            "standard_scores": {
                "package": {
                    "bucket": "bulk-ESG",
                    "name": "Bulk-ESG-Global-Scores-Wealth-Standard-v1",
                    "download": {
                        "path": "./downloads/esg/standard_scores",
                        "auto-retry": {"enabled": True, "count": 3},
                        "auto-extract": True,
                    },
                },
                "db": {
                    "connection": {
                        "module": "mysql",
                        "parameters": {
                            "user": "",
                            "password": "",
                            "host": "",
                            "port": 3306,
                            "database": "",
                        },
                    },
                    "table-auto-creation": True,
                    "create-table-queries": [
                        {
                            "table-name": "MY_STANDARD_SCORES_TABLE",
                            "create-query": "CREATE TABLE IF NOT EXISTS MY_STANDARD_SCORES_TABLE(ORGANIZATION varchar(255),INSERT_DATE varchar(255), INSTRUMENT varchar(255), PERIOD_END_DATE varchar(255), ESG_C_SCORE varchar(255),ESG_SCORE varchar(255),ENVIRONMENT_PILLAR_SCORE varchar(255), SOCIAL_PILLAR_SCORE varchar(255), GOVERNANCE_PILLAR_SCORE varchar(255), CONSTRAINT UC_MY_STANDARD_SCORES UNIQUE (INSTRUMENT,PERIOD_END_DATE))",
                            "check-if-exists-query": "SHOW TABLES",
                        }
                    ],
                    "search-query": "SELECT * FROM MY_STANDARD_SCORES_TABLE WHERE INSTRUMENT IN #{universe}",
                    "insert-queries": [
                        "REPLACE INTO MY_STANDARD_SCORES_TABLE(ORGANIZATION, INSERT_DATE, INSTRUMENT, PERIOD_END_DATE, ESG_C_SCORE, ESG_SCORE, ENVIRONMENT_PILLAR_SCORE, SOCIAL_PILLAR_SCORE, GOVERNANCE_PILLAR_SCORE) VALUES ('#{Fields.ESGOrganization.Names.Name.OrganizationName.0.OrganizationNormalizedName}', '#{File.DateTime}', '#{Fields.StatementDetails.OrganizationId}', '#{Fields.StatementDetails.FinancialPeriodEndDate}', '#{Fields.ESGScores.ESGCombinedScore.Value}', '#{Fields.ESGScores.ESGScore.Value}', '#{Fields.ESGScores.EnvironmentPillarScore.Value}', '#{Fields.ESGScores.SocialPillarScore.Value}', '#{Fields.ESGScores.GovernancePillarScore.Value}')"
                    ],
                    "stop-on-error": False,
                    "clean-up-queries": ["DELETE FROM MY_STANDARD_SCORES_TABLE"],
                    "output-fields-mapping": {
                        "INSTRUMENT": "instrument",
                        "PERIOD_END_DATE": "periodenddate",
                        "ESG_C_SCORE": "TR.TRESGCScore",
                        "ESG_SCORE": "TR.TRESGScore",
                        "ENVIRONMENT_PILLAR_SCORE": "TR.EnvironmentPillarScore",
                        "SOCIAL_PILLAR_SCORE": "TR.SocialPillarScore",
                        "GOVERNANCE_PILLAR_SCORE": "TR.GovernancePillarScore",
                    },
                },
            },
        },
    },
}

current_checksum = hashlib.md5(repr(data).encode()).hexdigest()  # NOSONAR
fixed_checksum = "08c37da0c079d1f55c88aca4e5c16287"

if current_checksum != fixed_checksum:
    warnings.warn(
        "Default library config was changed. This may cause unexpected errors. "
        "Please use user config to introduce new changes. You can reinstall "
        "refinitiv-data to revert changes back."
    )

config = config_from_dict(data)

# Run this module as script to update checksum after defaults change
if __name__ == "__main__":
    print("current checksum: ", current_checksum)



# coding: utf8

from .._object_definition import ObjectDefinition
from .._enums import AmericanMonteCarloMethod


class AmericanMonteCarloParameters(ObjectDefinition):
    def __init__(
        self,
        american_monte_carlo_method=None,
        additional_points=None,
        all_the_time_points_per_year=None,
        iteration_number=None,
    ):
        super().__init__()
        self.american_monte_carlo_method = american_monte_carlo_method
        self.additional_points = additional_points
        self.all_the_time_points_per_year = all_the_time_points_per_year
        self.iteration_number = iteration_number

    @property
    def american_monte_carlo_method(self):
        """
        :return: enum AmericanMonteCarloMethod
        """
        return self._get_enum_parameter(
            AmericanMonteCarloMethod, "americanMonteCarloMethod"
        )

    @american_monte_carlo_method.setter
    def american_monte_carlo_method(self, value):
        self._set_enum_parameter(
            AmericanMonteCarloMethod, "americanMonteCarloMethod", value
        )

    @property
    def additional_points(self):
        """
        :return: int
        """
        return self._get_parameter("additionalPoints")

    @additional_points.setter
    def additional_points(self, value):
        self._set_parameter("additionalPoints", value)

    @property
    def all_the_time_points_per_year(self):
        """
        :return: int
        """
        return self._get_parameter("allTheTimePointsPerYear")

    @all_the_time_points_per_year.setter
    def all_the_time_points_per_year(self, value):
        self._set_parameter("allTheTimePointsPerYear", value)

    @property
    def iteration_number(self):
        """
        :return: int
        """
        return self._get_parameter("iterationNumber")

    @iteration_number.setter
    def iteration_number(self, value):
        self._set_parameter("iterationNumber", value)



from refinitiv.data.content.ipa._curves import ZcCurveDefinition
from refinitiv.data.content.ipa._curves._cross_currency_curves import (
    FxForwardCurveDefinition,
    CrossCurrencyCurveDefinitionPricing,
    FxShiftScenario,
    ShiftDefinition,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    FxForwardCurveParameters,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._curves._request import (
    RequestItem,
)
from refinitiv.data.content.ipa._enums._extrapolation_mode import ExtrapolationMode
from refinitiv.data.content.ipa._enums._interpolation_mode import InterpolationMode
from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    ValuationTime,
)


def test_fx_forward_curves_request_item():
    # given
    expected_curve_definition = {
        "baseCurrency": "string",
        "baseIndexName": "string",
        "crossCurrencyDefinitions": [{}],
        "curveTenors": ["string1", "string2"],
        "interestRateCurveDefinitions": [{}],
        "isNonDeliverable": True,
        "pivotCurrency": "string",
        "pivotIndexName": "string",
        "quotedCurrency": "string",
        "quotedIndexName": "string",
    }

    curve_definition = FxForwardCurveDefinition(
        cross_currency_definitions=[CrossCurrencyCurveDefinitionPricing()],
        curve_tenors=["string1", "string2"],
        interest_rate_curve_definitions=[ZcCurveDefinition()],
        base_currency="string",
        base_index_name="string",
        is_non_deliverable=True,
        pivot_currency="string",
        pivot_index_name="string",
        quoted_currency="string",
        quoted_index_name="string",
    )

    expected_curve_parameters = {
        "extrapolationMode": "Constant",
        "ignoreInvalidInstrument": True,
        "ignorePivotCurrencyHolidays": True,
        "interpolationMode": "CubicDiscount",
        "useDelayedDataIfDenied": True,
        "valuationDate": "string",
        "valuationDateTime": "string",
        "valuationTime": {},
    }

    # when
    curve_parameters = FxForwardCurveParameters(
        extrapolation_mode=ExtrapolationMode.CONSTANT,
        interpolation_mode=InterpolationMode.CUBIC_DISCOUNT,
        valuation_time=ValuationTime(),
        ignore_invalid_instrument=True,
        ignore_pivot_currency_holidays=True,
        use_delayed_data_if_denied=True,
        valuation_date="string",
        valuation_date_time="string",
    )

    expected_shift_scenarios = {"fxCurveShift": {}, "shiftTag": "string"}

    # when
    shift_scenarios = FxShiftScenario(
        fx_curve_shift=ShiftDefinition(),
        shift_tag="string",
    )

    # when
    testing_obj = RequestItem(
        curve_definition=curve_definition,
        curve_parameters=curve_parameters,
        shift_scenarios=[shift_scenarios],
        curve_tag="curve_tag",
    )

    # then
    assert testing_obj.get_dict() == {
        "curveDefinition": expected_curve_definition,
        "curveParameters": expected_curve_parameters,
        "shiftScenarios": [expected_shift_scenarios],
        "curveTag": "curve_tag",
    }

    assert testing_obj.curve_tag == "curve_tag"



# coding: utf8

from enum import Enum, unique


@unique
class IndexAverageMethod(Enum):
    ARITHMETIC_AVERAGE = "ArithmeticAverage"
    COMPOUNDED_ACTUAL = "CompoundedActual"
    COMPOUNDED_AVERAGE_RATE = "CompoundedAverageRate"
    DAILY_COMPOUNDED_AVERAGE = "DailyCompoundedAverage"



from typing import Optional, List

from ._butterfly_shift import ButterflyShift
from ._combined_shift import CombinedShift
from ._flattening_shift import FlatteningShift
from ._long_end_shift import LongEndShift
from ._parallel_shift import ParallelShift
from ._short_end_shift import ShortEndShift
from ._time_bucket_shift import TimeBucketShift
from ._twist_shift import TwistShift
from ...._object_definition import ObjectDefinition
from ......_tools import try_copy_to_list


class ShiftDefinition(ObjectDefinition):
    """
    Generates the Cross Currency curves for the definitions provided

    Parameters
    ----------
    butterfly_shift : ButterflyShift, optional
        The definition of attributes for butterfly shift scenario. the start and end
        curve points are shifted by amount. the curve point at pivot tenor is not
        shifted.
    combined_shifts : list of CombinedShift, optional
        The definition of attributes for combined shifts scenario. each element of the
        combined shifts scenario describes separate shift scenario applied to curve
        points. if one scenario follows the other, they are added to each other.
    flattening_shift : FlatteningShift, optional
        The definition of attributes for flattening / steepening shift scenario. the
        start curve point is shifted by [amount / 2], the end curve point is shifted by
        [-amount / 2].
    long_end_shift : LongEndShift, optional
        The definition of attributes for long end shift scenario. the end curve point is
        shifted by amount. the start curve point is not shifted. the shift rises from 0
        to amount between the start and the end curve point.
    parallel_shift : ParallelShift, optional
        The definition of attributes for parallel shift scenario. each curve point is
        shifted by amount.
    short_end_shift : ShortEndShift, optional
        The definition of attributes for short end shift scenario. the start curve point
        is shifted by amount. the end curve point is not shifted. the shift decreases
        from 0 to amount between the start and the end curve point
    time_bucket_shifts : list of TimeBucketShift, optional
        The definition of attributes for time bucket shift scenario. each element of
        timebucketshifts describes the separate parallel shift. this shift applies from
        starttenor to endtenor.
    twist_shift : TwistShift, optional
        The definition of attributes for twist shift scenario. the start and end curve
        points are shifted by amount. the curve point at pivot tenor is not shifted.
    """

    def __init__(
        self,
        butterfly_shift: Optional[ButterflyShift] = None,
        combined_shifts: Optional[List[CombinedShift]] = None,
        flattening_shift: Optional[FlatteningShift] = None,
        long_end_shift: Optional[LongEndShift] = None,
        parallel_shift: Optional[ParallelShift] = None,
        short_end_shift: Optional[ShortEndShift] = None,
        time_bucket_shifts: Optional[List[TimeBucketShift]] = None,
        twist_shift: Optional[TwistShift] = None,
    ) -> None:
        super().__init__()
        self.butterfly_shift = butterfly_shift
        self.combined_shifts = try_copy_to_list(combined_shifts)
        self.flattening_shift = flattening_shift
        self.long_end_shift = long_end_shift
        self.parallel_shift = parallel_shift
        self.short_end_shift = short_end_shift
        self.time_bucket_shifts = try_copy_to_list(time_bucket_shifts)
        self.twist_shift = twist_shift

    @property
    def butterfly_shift(self):
        """
        The definition of attributes for butterfly shift scenario. the start and end
        curve points are shifted by amount. the curve point at pivot tenor is not
        shifted.
        :return: object ButterflyShift
        """
        return self._get_object_parameter(ButterflyShift, "butterflyShift")

    @butterfly_shift.setter
    def butterfly_shift(self, value):
        self._set_object_parameter(ButterflyShift, "butterflyShift", value)

    @property
    def combined_shifts(self):
        """
        The definition of attributes for combined shifts scenario. each element of the
        combined shifts scenario describes separate shift scenario applied to curve
        points. if one scenario follows the other, they are added to each other.
        :return: list CombinedShift
        """
        return self._get_list_parameter(CombinedShift, "combinedShifts")

    @combined_shifts.setter
    def combined_shifts(self, value):
        self._set_list_parameter(CombinedShift, "combinedShifts", value)

    @property
    def flattening_shift(self):
        """
        The definition of attributes for flattening / steepening shift scenario. the
        start curve point is shifted by [amount / 2], the end curve point is shifted by
        [-amount / 2].
        :return: object FlatteningShift
        """
        return self._get_object_parameter(FlatteningShift, "flatteningShift")

    @flattening_shift.setter
    def flattening_shift(self, value):
        self._set_object_parameter(FlatteningShift, "flatteningShift", value)

    @property
    def long_end_shift(self):
        """
        The definition of attributes for long end shift scenario. the end curve point is
        shifted by amount. the start curve point is not shifted. the shift rises from 0
        to amount between the start and the end curve point.
        :return: object LongEndShift
        """
        return self._get_object_parameter(LongEndShift, "longEndShift")

    @long_end_shift.setter
    def long_end_shift(self, value):
        self._set_object_parameter(LongEndShift, "longEndShift", value)

    @property
    def parallel_shift(self):
        """
        The definition of attributes for parallel shift scenario. each curve point is
        shifted by amount.
        :return: object ParallelShift
        """
        return self._get_object_parameter(ParallelShift, "parallelShift")

    @parallel_shift.setter
    def parallel_shift(self, value):
        self._set_object_parameter(ParallelShift, "parallelShift", value)

    @property
    def short_end_shift(self):
        """
        The definition of attributes for short end shift scenario. the start curve point
        is shifted by amount. the end curve point is not shifted. the shift decreases
        from 0 to amount between the start and the end curve point
        :return: object ShortEndShift
        """
        return self._get_object_parameter(ShortEndShift, "shortEndShift")

    @short_end_shift.setter
    def short_end_shift(self, value):
        self._set_object_parameter(ShortEndShift, "shortEndShift", value)

    @property
    def time_bucket_shifts(self):
        """
        The definition of attributes for time bucket shift scenario. each element of
        timebucketshifts describes the separate parallel shift. this shift applies from
        starttenor to endtenor.
        :return: list TimeBucketShift
        """
        return self._get_list_parameter(TimeBucketShift, "timeBucketShifts")

    @time_bucket_shifts.setter
    def time_bucket_shifts(self, value):
        self._set_list_parameter(TimeBucketShift, "timeBucketShifts", value)

    @property
    def twist_shift(self):
        """
        The definition of attributes for twist shift scenario. the start and end curve
        points are shifted by amount. the curve point at pivot tenor is not shifted.
        :return: object TwistShift
        """
        return self._get_object_parameter(TwistShift, "twistShift")

    @twist_shift.setter
    def twist_shift(self, value):
        self._set_object_parameter(TwistShift, "twistShift", value)



from typing import Union, List, Dict

from ._context import UDFMixin
from ._hp_context import HPContext


class HPUDFContext(UDFMixin, HPContext):
    def prepare_headers(self, raw: Union[list, dict], *args) -> List[List[Dict]]:
        headers = [{"displayName": "Instrument"}, {"displayName": "Date"}]

        if isinstance(raw, list):
            raw = raw[0]["headers"]

        elif isinstance(raw, dict):
            raw = raw["headers"]

        else:
            raw = []

        for item in raw:
            name = item["name"]

            if name.lower() in {"date", "instrument"}:
                continue

            headers.append({"displayName": name, "field": name})

        return [headers]



from refinitiv.data import _configure as configure
from refinitiv.data import _config_defaults


def test_defaults():
    assert configure.defaults


def test_http_request_timeout():
    assert configure.defaults.http_request_timeout


def test_log_level():
    assert configure.defaults.log_level


def test_platform_server_mode():
    assert configure.defaults.platform_server_mode is not None


def test_checksum():
    assert _config_defaults.current_checksum == _config_defaults.fixed_checksum



from functools import partial
from unittest import mock

import pytest

import refinitiv.data as rd
from refinitiv.data import OpenState
from refinitiv.data._content_type import ContentType
from refinitiv.data.content._universe_streams import _UniverseStreams
from refinitiv.data.content.custom_instruments._stream_facade import (
    CustomInstsUniverseStreams,
    Stream as CustomInstrumentsStreamFacade,
)
from refinitiv.data.content.ipa.financial_contracts._quantitative_data_stream import (
    QuantitativeDataStream,
)
from refinitiv.data.content.ipa.financial_contracts._stream_facade import (
    Stream as FinancialContractsStreamFacade,
)
from refinitiv.data.content.pricing._stream_facade import Stream as PricingStreamFacade
from refinitiv.data.content.pricing.chain._stream import StreamingChain
from refinitiv.data.content.pricing.chain._stream_facade import (
    Stream as PricingChainStreamFacade,
)
from refinitiv.data.content.trade_data_service._stream import TradeDataStream
from refinitiv.data.content.trade_data_service._stream_facade import (
    Stream as TradeDataStreamFacade,
)
from refinitiv.data.delivery._stream import (
    _RDPStream,
    StreamState,
    _OMMStream,
    RDPStream,
    OMMStream,
)
from tests.unit.conftest import StubSession, StubResponse

ids = [
    "pricing",
    "trade_data_service",
    "rdp_stream",
    "omm_stream",
    "pricing.chain",
    "financial_contracts.bond",
    "custom_instruments",
]


@pytest.mark.parametrize(
    argnames="definition",
    argvalues=[
        rd.content.pricing.Definition("universe"),
        rd.content.trade_data_service.Definition(),
        rd.delivery.rdp_stream.Definition("", [], [], {}, "api/path"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.pricing.chain.Definition("universe"),
        rd.content.ipa.financial_contracts.bond.Definition("universe"),
        rd.content.custom_instruments.Definition("universe"),
    ],
    ids=ids,
)
def test_stream_can_be_opened_via_context_manager(definition):
    session = StubSession(is_open=True)
    stream = definition.get_stream(session)
    try:
        stream._stream._stream._do_open = lambda *_, **__: None
        stream._stream._stream._do_close = lambda *_, **__: None
    except AttributeError:
        stream._stream._do_open = lambda *_, **__: None
        stream._stream._do_close = lambda *_, **__: None

    with stream:
        assert stream.open_state is OpenState.Opened

    assert stream.open_state is OpenState.Closed


def test_rdp_stream_got_alarm_message():
    # given
    def on_error(*args, **kwargs):
        nonlocal error_printed
        error_printed = True

    session = StubSession()
    error_printed = False
    stream = _RDPStream(3, session, "", [], [], {}, {})
    stream._error = on_error
    stream._state = StreamState.Opening
    alarm_message = {
        "data": [],
        "state": {
            "id": "QPSValuation.ERROR_REQUEST_TIMEOUT",
            "code": 408,
            "status": "ERROR",
            "message": "The request could not be executed within the service allocated time",
            "stream": "Open",
        },
        "type": "Alarm",
        "streamID": "3",
    }

    # when
    stream.dispatch_alarm({}, alarm_message)

    # then
    assert error_printed is True
    assert stream._opened.is_set() is True


@pytest.mark.parametrize(
    argnames="definition",
    argvalues=[
        rd.content.pricing.Definition("universe"),
        rd.content.trade_data_service.Definition(),
        rd.delivery.rdp_stream.Definition("", [], [], {}, "api/path"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.pricing.chain.Definition("universe"),
        rd.content.ipa.financial_contracts.bond.Definition("universe"),
        rd.content.custom_instruments.Definition("universe"),
    ],
    ids=ids,
)
def test_stream_open_always_use_default_session_if_not_passed_in_get_stream(definition):
    # given
    session_a = StubSession(is_open=True)
    rd.session.set_default(session_a)
    stream = definition.get_stream()
    if hasattr(stream._stream, "_stream"):
        stream._stream._stream._do_open = lambda *args, **kwargs: None
    else:
        stream._stream._do_open = lambda *args, **kwargs: None

    assert stream._stream.session == session_a

    # when
    session_b = StubSession(is_open=True)
    rd.session.set_default(session_b)
    stream.open()

    # then
    assert stream._stream.session == session_b

    rd.session.set_default(None)


@pytest.mark.parametrize(
    argnames="definition",
    argvalues=[
        rd.content.pricing.Definition("universe"),
        rd.content.trade_data_service.Definition(),
        rd.delivery.rdp_stream.Definition("", [], [], {}, "api/path"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.pricing.chain.Definition("universe"),
        rd.content.ipa.financial_contracts.bond.Definition("universe"),
        rd.content.custom_instruments.Definition("universe"),
    ],
    ids=ids,
)
def test_stream_open_always_use_session_from_get_stream_if_passed(definition):
    # given
    session_a = StubSession(is_open=True)
    rd.session.set_default(session_a)
    stream = definition.get_stream(session_a)
    if hasattr(stream._stream, "_stream"):
        stream._stream._stream._do_open = lambda *args, **kwargs: None
    else:
        stream._stream._do_open = lambda *args, **kwargs: None

    assert stream._stream.session == session_a

    # when
    session_b = StubSession(is_open=True)
    rd.session.set_default(session_b)
    stream.open()

    # then
    assert stream._stream.session == session_a

    rd.session.set_default(None)


@pytest.mark.parametrize(
    argnames="definition",
    argvalues=[
        rd.content.pricing.Definition("universe"),
        rd.content.trade_data_service.Definition(),
        rd.delivery.rdp_stream.Definition("", [], [], {}, "api/path"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.pricing.chain.Definition("universe"),
        rd.content.ipa.financial_contracts.bond.Definition("universe"),
        rd.content.custom_instruments.Definition("universe"),
    ],
    ids=ids,
)
def test_user_cannot_create_stream_without_any_session(definition):
    # then
    with pytest.raises(
        AttributeError,
        match="No default session created yet. Please create a session first!",
    ):
        # given
        definition.get_stream()


@pytest.mark.parametrize(
    argnames="definition",
    argvalues=[
        rd.content.pricing.Definition("universe"),
        rd.content.trade_data_service.Definition(),
        rd.delivery.rdp_stream.Definition("", [], [], {}, "api/path"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.pricing.chain.Definition("universe"),
        rd.content.ipa.financial_contracts.bond.Definition("universe"),
        rd.content.custom_instruments.Definition("universe"),
    ],
    ids=ids,
)
def test_user_can_create_stream_with_opened_default_session(definition):
    # given
    session = StubSession(is_open=True)
    rd.session.set_default(session)

    # when
    try:
        definition.get_stream()
    except Exception as e:
        assert False, str(e)
    else:
        # then
        assert True

    rd.session.set_default(None)


@pytest.mark.parametrize(
    argnames="definition",
    argvalues=[
        rd.content.pricing.Definition("universe"),
        rd.content.trade_data_service.Definition(),
        rd.delivery.rdp_stream.Definition("", [], [], {}, "api/path"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.pricing.chain.Definition("universe"),
        rd.content.ipa.financial_contracts.bond.Definition("universe"),
    ],
    ids=[
        "pricing",
        "trade_data_service",
        "rdp_stream",
        "omm_stream",
        "pricing.chain",
        "financial_contracts.bond",
    ],
)
def test_user_can_create_stream_with_not_opened_default_session(definition):
    # given
    session = StubSession()
    rd.session.set_default(session)

    try:
        # when
        definition.get_stream()
    except Exception as e:
        assert False, str(e)
    else:
        # then
        assert True

    rd.session.set_default(None)


@pytest.mark.parametrize(
    argnames="definition",
    argvalues=[
        rd.content.custom_instruments.Definition("universe"),
    ],
    ids=[
        "custom_instruments",
    ],
)
def test_user_cannot_create_stream_with_not_opened_default_session(definition):
    # given
    session = StubSession()
    rd.session.set_default(session)

    # then
    with pytest.raises(
        ValueError, match="Session is not opened. Can't send any request"
    ):
        # when
        definition.get_stream()

    rd.session.set_default(None)


@pytest.mark.parametrize(
    argnames="internal_stream_class",
    argvalues=[
        partial(_UniverseStreams, ContentType.STREAMING_CUSTOM, "universe"),
        TradeDataStream,
        partial(
            _RDPStream,
            **{
                "stream_id": 0,
                "service": "",
                "universe": [],
                "view": [],
                "parameters": {},
                "details": {},
            },
        ),
        partial(_OMMStream, **{"stream_id": 0, "name": "", "details": {}}),
        partial(StreamingChain, **{"name": ""}),
        partial(QuantitativeDataStream, **{"universe": ""}),
        partial(CustomInstsUniverseStreams, **{"content_type": {}, "universe": ""}),
    ],
    ids=ids,
)
def test_internal_stream_cannot_be_created_without_session(internal_stream_class):
    # then
    with pytest.raises(
        AttributeError, match="'NoneType' object has no attribute 'logger'"
    ):
        # when
        internal_stream_class(session=None)


@pytest.mark.parametrize(
    argnames="facade_stream_class",
    argvalues=[
        PricingStreamFacade,
        TradeDataStreamFacade,
        partial(
            RDPStream,
            **{
                "api": "",
                "service": "",
                "universe": [],
                "view": [],
                "parameters": {},
            },
        ),
        partial(OMMStream, **{"name": ""}),
        partial(PricingChainStreamFacade, **{"name": ""}),
        partial(FinancialContractsStreamFacade, **{"universe": ""}),
        partial(CustomInstrumentsStreamFacade, **{"uuid": "uuid"}),
    ],
    ids=ids,
)
def test_facade_stream_can_be_created_with_not_opened_default_session(
    facade_stream_class,
):
    # given
    session = StubSession()
    rd.session.set_default(session)

    try:
        # when
        facade_stream_class()
    except Exception as e:
        assert False, str(e)
    else:
        # then
        assert True

    rd.session.set_default(None)


@pytest.mark.parametrize(
    argnames="facade_stream_class",
    argvalues=[
        CustomInstrumentsStreamFacade,
    ],
    ids=[
        "custom_instruments",
    ],
)
def test_facade_stream_cannot_be_created_with_not_opened_default_session(
    facade_stream_class,
):
    # given
    session = StubSession()
    rd.session.set_default(session)

    # then
    with pytest.raises(
        ValueError, match="Session is not opened. Can't send any request"
    ):
        # when
        facade_stream_class()

    rd.session.set_default(None)


@pytest.mark.parametrize(
    argnames="internal_stream_class",
    argvalues=[
        partial(
            _UniverseStreams,
            **{
                "content_type": ContentType.STREAMING_CUSTOM,
                "universe": "universe",
                "api": "api",
            },
        ),
        TradeDataStream,
        partial(
            _RDPStream,
            **{
                "stream_id": 0,
                "service": "",
                "universe": [],
                "view": [],
                "parameters": {},
                "details": {},
            },
        ),
        partial(_OMMStream, **{"stream_id": 0, "name": "", "details": {}}),
        partial(StreamingChain, **{"name": ""}),
        partial(QuantitativeDataStream, **{"universe": ""}),
        partial(CustomInstsUniverseStreams, **{"content_type": {}, "universe": ""}),
    ],
    ids=ids,
)
def test_internal_stream_cannot_set_new_session_if_stream_already_opened(
    internal_stream_class,
):
    # given
    session_first = StubSession(is_open=True)
    stream = internal_stream_class(session=session_first)
    if hasattr(stream, "_stream"):
        stream._stream._do_open = lambda *args, **kwargs: None
    else:
        stream._do_open = lambda *args, **kwargs: None

    # when
    stream.open()
    session_second = StubSession(is_open=True)
    stream.session = session_second

    # then
    assert stream.session == session_first


class StubOMMStream:
    def __init__(self, service, *args, **kwargs):
        self.classname = ...
        self.on = lambda *args, **kwargs: ...
        self.open = lambda *args, **kwargs: ...
        self.id = ...
        assert service == "test_service"


infos = {
    "services": [
        {
            "port": 443,
            "location": ["ap-northeast-1a"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "ap-northeast-1-aws-1-sm.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
        },
    ]
}


def http_request(request):
    url = request.url
    if url.startswith("test_get_rdp_url_root/streaming"):
        return StubResponse(content_data=infos)
    return StubResponse({})


@pytest.mark.parametrize(
    ("name_in_config", "definition"),
    [
        ("pricing", rd.content.pricing.Definition("EUR=")),
        ("custom-instruments", rd.content.custom_instruments.Definition("EUR=")),
    ],
)
def test_service_in_config_omm(name_in_config, definition):
    session = StubSession(is_open=True, stream_auto_reconnection=True)
    session.http_request = http_request
    session.config.set_param(
        f"apis.streaming.{name_in_config}.service", "test_service", auto_create=True
    )
    with mock.patch(
        "refinitiv.data.delivery._stream._stream_factory._OMMStream", new=StubOMMStream
    ):
        stream = definition.get_stream(session=session)
        stream.open()


class StubRDPStream:
    def __init__(self, service, *args, **kwargs):
        self.classname = ...
        self.on_response = lambda *args, **kwargs: ...
        self.on_update = lambda *args, **kwargs: ...
        self.on_ack = lambda *args, **kwargs: ...
        self.on_alarm = lambda *args, **kwargs: ...
        self.on = lambda *args, **kwargs: ...
        self.open = lambda *args, **kwargs: ...
        assert service == "test_service"


@pytest.mark.parametrize(
    ("name_in_config", "definition"),
    [
        ("trading-analytics", rd.content.trade_data_service.Definition()),
        (
            "quantitative-analytics",
            rd.content.ipa.financial_contracts.option.Definition(),
        ),
    ],
)
def test_service_in_config_rdp(name_in_config, definition):
    session = StubSession(is_open=True, stream_auto_reconnection=True)
    session.http_request = http_request
    session.config.set_param(
        f"apis.streaming.{name_in_config}.service", "test_service", auto_create=True
    )
    with mock.patch(
        "refinitiv.data.delivery._stream._stream_factory._RDPStream", new=StubRDPStream
    ):
        stream = definition.get_stream(session=session)
        stream.open()


{"apis": {"streaming": {"pricing": {"service": "MY_SERVICE"}}}}



from refinitiv.data._tools import cached_property


class StreamStateEvent:
    OPENING = "opening_state_event"
    OPENED = "opened_state_event"
    CLOSING = "closing_state_event"
    CLOSED = "closed_state_event"


class StreamEvent:
    # broadcast events
    UPDATE = "update_stream_event"
    REFRESH = "refresh_stream_event"
    STATUS = "status_stream_event"
    COMPLETE = "complete_stream_event"
    ERROR = "error_stream_event"
    ACK = "ack_stream_event"
    RESPONSE = "response_stream_event"
    ALARM = "alarm_stream_event"

    # specific events
    _UPDATE = "update_stream_{}_event"
    _REFRESH = "refresh_stream_{}_event"
    _STATUS = "status_stream_{}_event"
    _COMPLETE = "complete_stream_{}_event"
    _ERROR = "error_stream_{}_event"
    _ACK = "ack_stream_{}_event"
    _RESPONSE = "response_stream_{}_event"
    _ALARM = "alarm_stream_{}_event"

    _cache = {}

    def __init__(self, stream_id) -> None:
        self.stream_id = stream_id

    @cached_property
    def update_by_id(self) -> str:
        return self._UPDATE.format(self.stream_id)

    @cached_property
    def refresh_by_id(self) -> str:
        return self._REFRESH.format(self.stream_id)

    @cached_property
    def status_by_id(self) -> str:
        return self._STATUS.format(self.stream_id)

    @cached_property
    def complete_by_id(self) -> str:
        return self._COMPLETE.format(self.stream_id)

    @cached_property
    def error_by_id(self) -> str:
        return self._ERROR.format(self.stream_id)

    @cached_property
    def ack_by_id(self) -> str:
        return self._ACK.format(self.stream_id)

    @cached_property
    def response_by_id(self) -> str:
        return self._RESPONSE.format(self.stream_id)

    @cached_property
    def alarm_by_id(self) -> str:
        return self._ALARM.format(self.stream_id)

    @classmethod
    def get(cls, stream_id: int) -> "StreamEvent":
        return cls._cache.setdefault(stream_id, StreamEvent(stream_id))


class StreamCxnEvent(object):
    # broadcast events
    CONNECTING = "connecting_cxn_event"
    CONNECTED = "connected_cxn_event"
    READY = "ready_cxn_event"
    DISCONNECTING = "disconnecting_cxn_event"
    DISCONNECTED = "disconnected_cxn_event"
    DISPOSED = "disposed_cxn_event"
    RECONNECTING = "reconnecting_cxn_event"
    RECONNECTED = "reconnected_cxn_event"
